"""Configuration and tasks"""

import json

from attr import define
import fasteners

from experimaestro import taskglobals

from termcolor import cprint
from pathlib import Path
import logging
import io
from enum import Enum
import inspect
import importlib
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    overload,
    TYPE_CHECKING,
)
import sys
import experimaestro
from experimaestro.utils import logger
from experimaestro.core.types import DeprecatedAttribute, ObjectType, TypeVarType
from ..context import SerializationContext, SerializedPath, SerializedPathLoader

if TYPE_CHECKING:
    from ..callbacks import TaskEventListener
    from ..identifier import Identifier
    from experimaestro.scheduler.base import Job
    from experimaestro.scheduler.workspace import RunMode
    from experimaestro.launchers import Launcher
    from experimaestro.scheduler import Workspace

from .config_walk import ConfigWalk, ConfigWalkContext
from .config_utils import (
    getqualattr,
    add_to_path,
    TaggedValue,
    ObjectStore,
    classproperty,
)

T = TypeVar("T", bound="Config")


DependentMarker = Callable[["Config"], None]


def updatedependencies(
    dependencies, value: "Config", path: List[str], taskids: Set[int]
):
    """Search recursively jobs to add them as dependencies

    Arguments:
        dependencies: The current set of dependencies
        value: The current inspected configuration
        path: The current path (for error tracing)
        taskids: Sets of added tasks (ids) to avoid repeated depencies
    """

    # Adds pre-tasks
    if isinstance(value, Config):
        value.__xpm__.updatedependencies(dependencies, path, taskids)
    elif isinstance(value, (list, set)):
        for el in value:
            updatedependencies(dependencies, el, path, taskids)
    elif isinstance(value, (dict,)):
        for key, val in value.items():
            updatedependencies(dependencies, key, path, taskids)
            updatedependencies(dependencies, val, path, taskids)
    elif isinstance(value, (str, int, float, Path, Enum)):
        pass
    else:
        raise NotImplementedError("update dependencies for type %s" % type(value))


NOT_SET = object()


@define()
class WatchedOutput:
    #: The configuration containing the watched output
    config: "ConfigInformation"

    #: The watched output (name)
    method_name: str

    #: The watched output method (called with the JSON event)
    method: Callable

    #: The callback to call (with the output of the previous method)
    callback: Callable

    #: The enclosing job (set when registered with scheduler)
    job: Optional["Job"] = None


def get_generated_paths(
    v: Union["ConfigMixin", list, dict],
    path: list[str] | None = None,
    paths: list[str] | None = None,
) -> list[str]:
    """Get the list of generated paths, useful to track down those

    :param path: The current path
    :param paths: The list of generated paths so far, defaults to None
    :return: The full list of generated paths
    """
    paths = [] if paths is None else paths
    path = [] if path is None else path

    if isinstance(v, list):
        for ix, element in enumerate(v):
            get_generated_paths(element, path + [f"[{ix}]"], paths)

    elif isinstance(v, dict):
        for key, element in v.items():
            get_generated_paths(element, path + [f"[{key}]"], paths)

    elif isinstance(v, ConfigMixin):
        for key in v.__xpm__._generated_values:
            value = v.__xpm__.values[key]
            if isinstance(value, ConfigMixin) and value.__xpm__._generated_values:
                path.append(key)
                get_generated_paths(value, path, paths)
                path.pop()
            else:
                paths.append(".".join(path + [key]))
    else:
        raise ValueError(f"Cannot handle type {type(v)}")
    return paths


class ConfigInformation:
    """Holds experimaestro information for a config (or task) instance"""

    _meta: Optional[bool]
    """Forces this configuration to be a meta-parameter"""

    # Set to true when loading from JSON
    LOADING: ClassVar[bool] = False

    def __init__(self, pyobject: "ConfigMixin"):
        # The underlying pyobject and XPM type
        self.pyobject = pyobject
        self.xpmtype: "ObjectType" = pyobject.__xpmtype__
        self.values = {}

        # Meta-informations
        self._tags: dict[str, Any] = {}
        self._initinfo = ""

        self._taskoutput = None
        """Task output (caches the value of a submit)"""

        self.task: Optional["Config"] = None
        """The task this configuration depends upon (or None)"""

        # State information
        self.job = None
        self._job_listener: "TaskEventListener" | None = None

        #: True when this configuration was loaded from disk
        self.loaded = False

        # Explicitly added dependencies
        self.dependencies = []

        # Concrete type variables resolutions
        # This is used to check typevars coherence
        self.concrete_typevars: Dict[TypeVar, type] = {}

        # Initialization tasks
        self.init_tasks: List["LightweightTask"] = []

        # Watched outputs
        self.watched_outputs: List[WatchedOutput] = []

        # Cached information

        self._identifier = None
        """The configuration identifier (cached when sealed)"""

        self._validated = False
        self._sealed = False
        self._meta = None

        # This contains the list of generated values (using context) in this
        # configuration or any sub-configuration, is generated. This prevents
        # problem when a configuration with generated values is re-used.
        self._generated_values = []

    def set_meta(self, value: Optional[bool]):
        """Sets the meta flag"""
        assert not self._sealed, "Configuration is sealed"
        self._meta = value

    @property
    def meta(self):
        return self._meta

    def get(self, name):
        """Get an XPM managed value"""
        if name in self.xpmtype.arguments:
            return self.values[name]

        # Not an argument, bypass
        return object.__getattribute__(self.pyobject, name)

    @staticmethod
    def is_generated_value(argument, value):
        if argument.ignore_generated:
            return False

        if value is None:
            return False

        if isinstance(value, (int, str, float, bool, Enum, Path)):
            return False

        if isinstance(value, ConfigMixin):
            return value.__xpm__._generated_values and value.__xpm__.task is None

        if isinstance(value, list):
            return any(ConfigInformation.is_generated_value(argument, x) for x in value)

        if isinstance(value, dict):
            return any(
                ConfigInformation.is_generated_value(argument, x)
                for x in value.values()
            )

        return False

    def set(self, k, v, bypass=False):
        from experimaestro.generators import Generator

        # Not an argument, bypass
        if k not in self.xpmtype.arguments:
            setattr(self.pyobject, k, v)
            return

        if self._sealed and not bypass:
            raise AttributeError(f"Object is read-only (trying to set {k})")

        if not isinstance(v, ConfigMixin) and isinstance(v, Config):
            raise AttributeError(
                "Configuration (and not objects) should be used. Consider using .C(...)"
            )

        try:
            argument = self.xpmtype.arguments.get(k, None)
            if argument:
                if ConfigInformation.is_generated_value(argument, v):
                    raise AttributeError(
                        f"Cannot set {k} to a configuration with generated values. "
                        "Here is the list of paths to help you: "
                        f"""{', '.join(get_generated_paths(v, [k]))}"""
                    )

                if not bypass and (
                    (isinstance(argument.generator, Generator)) or argument.constant
                ):
                    raise AttributeError("Property %s is read-only" % (k))
                if v is not None:
                    self.values[k] = argument.validate(v)
                    # Check for type variables
                    if type(argument.type) is TypeVarType:
                        self.check_typevar(argument.type.typevar, type(v))
                    if isinstance(v, Config):
                        # If the value is a Config, fuse type variables
                        v.__xpm__.fuse_concrete_typevars(self.concrete_typevars)
                        self.fuse_concrete_typevars(v.__xpm__.concrete_typevars)
                elif argument.required:
                    raise AttributeError("Cannot set required attribute to None")
                else:
                    self.values[k] = None
            else:
                raise AttributeError(
                    "Cannot set non existing attribute %s in %s" % (k, self.xpmtype)
                )
        except Exception:
            logger.error("Error while setting value %s in %s", k, self.xpmtype)
            raise

    def fuse_concrete_typevars(self, typevars: Dict[TypeVar, type]):
        """Fuses concrete type variables with the current ones"""
        for typevar, v in typevars.items():
            self.check_typevar(typevar, v)

    def check_typevar(self, typevar: TypeVar, v: type):
        """Check if a type variable is coherent with the current typevars bindings,
        updates the bindings if necessary"""
        if typevar not in self.concrete_typevars:
            self.concrete_typevars[typevar] = v
            return

        concrete_typevar = self.concrete_typevars[typevar]
        bound = typevar.__bound__
        # Check that v is a subclass of the typevar OR that typevar is a subclass of v
        # Then set the concrete type variable to the most generic type

        # First, limiting to the specified bound
        if bound is not None:
            if not issubclass(v, bound):
                raise TypeError(
                    f"Type variable {typevar} is bound to {bound}, but tried to set it to {v}"
                )

        if issubclass(v, concrete_typevar):
            # v is a subclass of the typevar, keep the typevar
            return
        if issubclass(concrete_typevar, v):
            # typevar is a subclass of v, keep v
            self.concrete_typevars[typevar] = v
            return
        raise TypeError(
            f"Type variable {typevar} is already set to {self.concrete_typevars[typevar]}, "
            f"but tried to set it to {v}"
            f" (current typevars bindings: {self.concrete_typevars})"
        )

    def addtag(self, name, value):
        self._tags[name] = value

    def xpmvalues(self, generated=False):
        """Returns an iterarator over arguments and associated values"""
        for argument in self.xpmtype.arguments.values():
            if argument.name in self.values or (generated and argument.generator):
                yield argument, self.values[argument.name]

    def tags(self):
        class TagFinder(ConfigWalk):
            def __init__(self):
                super().__init__(recurse_task=True)
                self.tags = {}

            def postprocess(self, stub, config: Config, values):
                self.tags.update(config.__xpm__._tags)
                return self.tags

        return TagFinder()(self.pyobject)

    def validate(self):
        """Validate a value"""
        if not self._validated:
            self._validated = True

            # Check each argument
            for k, argument in self.xpmtype.arguments.items():
                value = self.values.get(k)
                if value is not None:
                    if isinstance(value, Config):
                        value.__xpm__.validate()
                elif argument.required:
                    if not argument.generator:
                        raise ValueError(
                            "Value %s is required but missing when building %s at %s"
                            % (k, self.xpmtype, self._initinfo)
                        )

            # Validate init tasks
            for init_task in self.init_tasks:
                init_task.__xpm__.validate()

            # Use __validate__ method
            if hasattr(self.pyobject, "__validate__"):
                try:
                    self.pyobject.__validate__()
                except Exception:
                    logger.error(
                        "Error while validating %s at %s", self.xpmtype, self._initinfo
                    )
                    raise

    def seal(self, context: ConfigWalkContext):
        """Seals the object and generate values when needed

        Arguments:
            - context: the generation context
        """
        if generated_keys := [
            k
            for k, v in self.values.items()
            if ConfigInformation.is_generated_value(self.xpmtype.arguments[k], v)
        ]:
            raise AttributeError(
                "Cannot seal a configuration with generated values:"
                f"""{",".join(generated_keys)} in {context.currentpath}"""
            )

        class Sealer(ConfigWalk):
            def preprocess(self, config: ConfigMixin):
                return not config.__xpm__._sealed, config

            def postprocess(self, stub, config: ConfigMixin, values):
                # Generate values
                from experimaestro.generators import Generator

                for k, argument in config.__xpmtype__.arguments.items():
                    try:
                        if argument.generator:
                            if not isinstance(argument.generator, Generator):
                                # Don't set if already set
                                if config.__xpm__.values.get(k) is not None:
                                    continue
                                value = argument.generator()
                            else:
                                # Generate a value
                                sig = inspect.signature(argument.generator)
                                if len(sig.parameters) == 0:
                                    value = argument.generator()
                                elif len(sig.parameters) == 2:
                                    # Only in that case do we need to flag this configuration
                                    # as containing generated values
                                    if not argument.ignore_generated:
                                        config.__xpm__._generated_values.append(k)
                                    else:
                                        logging.warning("Ignoring %s", k)
                                    value = argument.generator(self.context, config)
                                else:
                                    assert (
                                        False
                                    ), "generator has either two parameters (context and config), or none"
                            config.__xpm__.set(k, value, bypass=True)
                        else:
                            value = config.__xpm__.values.get(k)
                    except Exception:
                        logger.error(
                            "While setting %s of %s", argument.name, config.__xpmtype__
                        )
                        raise

                    # Propagate the generated value flag
                    if (
                        value is not None
                        and isinstance(value, ConfigMixin)
                        and value.__xpm__._generated_values
                    ):
                        if not argument.ignore_generated:
                            config.__xpm__._generated_values.append(k)
                        else:
                            logging.warning("Ignoring %s", k)

                config.__xpm__._sealed = True

        Sealer(context, recurse_task=True)(self.pyobject)

    def __unseal__(self):
        """Unseal this configuration and its descendant

        Internal API - do not use
        """
        context = ConfigWalkContext()

        class Unsealer(ConfigWalk):
            def preprocess(self, config: ConfigMixin):
                return config.__xpm__._sealed, config

            def postprocess(self, stub, config: ConfigMixin, values):
                config.__xpm__._sealed = False
                config.__xpm__._identifier = None

        Unsealer(context, recurse_task=True)(self.pyobject)

    @property
    def identifier(self):
        """Computes the unique identifier"""
        from ..identifier import IdentifierComputer

        # Computes raw identifier if needed
        if self._identifier is not None:
            return self._identifier

        # Get the main identifier
        identifier = IdentifierComputer.compute(self.pyobject)
        if self._sealed:
            self._identifier = identifier
        return identifier

    def dependency(self):
        """Returns a dependency"""
        from experimaestro.scheduler import JobDependency

        assert self.job, f"{self.xpmtype} is a task but was not submitted"
        return JobDependency(self.job)

    def updatedependencies(
        self,
        dependencies: Set["experimaestro.dependencies.Dependency"],
        path: List[str],
        taskids: Set[int],
    ):
        # Add initialization tasks
        for init_task in self.init_tasks:
            init_task.__xpm__.updatedependencies(
                dependencies, path + ["__init_tasks__"], taskids
            )

        # Check for an associated task (and not loaded)
        if self.task and not self.loaded:
            if id(self.task) not in taskids:
                taskids.add(id(self.task))
                dependencies.add(self.task.__xpm__.dependency())
        else:
            # Look arguments
            for argument, value in self.xpmvalues():
                try:
                    if value is not None:
                        updatedependencies(
                            dependencies, value, path + [argument.name], taskids
                        )
                except Exception:
                    logger.error("While setting %s", path + [argument.name])
                    raise

    def apply_submit_hooks(self, job: "Job", launcher: "Launcher"):
        """Apply configuration hooks"""
        context = ConfigWalkContext()

        class HookGatherer(ConfigWalk):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.hooks = set()

            def postprocess(self, stub, config: "Config", values: Dict[str, Any]):
                self.hooks.update(config.__xpmtype__.submit_hooks)

        gatherer = HookGatherer(context, recurse_task=False)
        gatherer(self.pyobject)

        # Apply hooks
        for hook in gatherer.hooks:
            hook.process(job, launcher)

    def validate_and_seal(self, context: ConfigWalkContext):
        try:
            self.validate()
        except Exception as e:
            logger.error(
                "Error while validating object of type %s, defined %s",
                self.xpmtype,
                self._initinfo,
            )
            raise e

        # Now, seal the object
        self.seal(context)

    def watch_output(self, method, callback):
        """Watch the task output linked with a given method

        :param method: The method to watch
        :param callback: The callback

        :raises TypeError: If the task is not a ResumableTask
        """
        # Only ResumableTask can have dynamic outputs - regular tasks
        # have their directories cleaned up, losing the output file
        if not isinstance(self.pyobject, ResumableTask):
            raise TypeError(
                f"Only ResumableTask can use watch_output. "
                f"{self.xpmtype} is not a ResumableTask. "
                "Dynamic outputs require the task directory to be preserved "
                "across restarts, which only ResumableTask provides."
            )

        watched = WatchedOutput(method.__self__, method.__name__, method, callback)
        self.watched_outputs.append(watched)
        if self.job:
            self.job.watch_output(watched)

    def on_completed(self, callback: Callable[[], None]):
        """Call a method when the task is completed successfully

        :param callback: _description_
        """
        from ..callbacks import TaskEventListener

        TaskEventListener.on_completed(self, callback)

    def submit(
        self,
        workspace: "Workspace",
        launcher: "Launcher",
        *,
        run_mode=None,
        init_tasks: List["LightweightTask"] = [],
        max_retries: Optional[int] = None,
    ):
        from experimaestro.scheduler import experiment, JobContext
        from experimaestro.scheduler.workspace import RunMode
        from ..callbacks import TaskEventListener

        # --- Prepare the object

        if self.job:
            raise Exception("task %s was already submitted" % self)
        if not self.xpmtype.task:
            raise ValueError("%s is not a task" % self.xpmtype)

        # --- Submit the job

        # Sets the init tasks
        self.init_tasks = init_tasks

        # Creates a new job
        self.job = self.xpmtype.task(
            self.pyobject,
            launcher=launcher,
            workspace=workspace,
            run_mode=run_mode,
            max_retries=max_retries,
        )

        # Validate the object
        job_context = JobContext(self.job)
        self.validate_and_seal(job_context)

        # --- Workspace

        workspace = workspace or (
            experiment.CURRENT.workspace if experiment.CURRENT else None
        )

        # --- Launcher

        launcher = (
            launcher
            or (workspace and workspace.launcher)
            or (experiment.CURRENT and experiment.CURRENT.workspace.launcher)
        )

        if launcher:
            launcher.onSubmit(self.job)

        # Apply submit hooks
        self.apply_submit_hooks(self.job, launcher)

        # Add job dependencies
        self.updatedependencies(self.job.dependencies, [], set([id(self.pyobject)]))

        # Add predefined dependencies
        self.job.dependencies.update(self.dependencies)

        run_mode = (
            workspace.run_mode if run_mode is None else run_mode
        ) or RunMode.NORMAL
        if run_mode == RunMode.NORMAL:
            TaskEventListener.connect(experiment.CURRENT)
            other = experiment.CURRENT.submit(self.job)
            if other:
                # Our job = previously submitted job
                self.job = other
        else:
            # Show a warning
            if run_mode == RunMode.GENERATE_ONLY:
                experiment.CURRENT.prepare(self.job)

            # Check if job is done
            tags = ", ".join(f"{k}={v}" for k, v in self.job.tags.items())
            s = f"""Simulating {self.job.relpath} {f"({tags})" if tags else ""}"""

            color = "white"
            if self.job.workspace is not None:
                if self.job.donepath.is_file():
                    color = "light_green"
                    cprint(f"[done] {s}", color, file=sys.stderr)
                elif self.job.failedpath.is_file():
                    color = "light_red"
                    cprint(f"[failed] {s}", color, file=sys.stderr)
                elif self.job.pidpath.is_file():
                    color = "blue"
                    cprint(f"[running] {s}", color, file=sys.stderr)
                else:
                    color = "light_blue"
                    cprint(f"[not run] {s}", color, file=sys.stderr)

                if launcher:
                    cprint(f"   [Launcher] {launcher}", color, file=sys.stderr)

                if not self.job.dependencies:
                    cprint("   [No dependencies]", color, file=sys.stderr)

                for dep in self.job.dependencies:
                    cprint(f"   [Dependency] {dep}", color, file=sys.stderr)

                print(file=sys.stderr)  # noqa: T201

        # Mark this configuration also
        self.task = self.pyobject

        if hasattr(self.pyobject, "task_outputs"):
            self._taskoutput = self.pyobject.task_outputs(self.mark_output)
        else:
            self._taskoutput = self.task = self.pyobject

        return self._taskoutput

    def mark_output(self, config: "Config"):
        """Sets a dependency on the job"""
        assert not isinstance(config, Task), "Cannot set a dependency on a task"
        assert isinstance(
            config, ConfigMixin
        ), "Only configurations can be marked as dependent on a task"
        config.__xpm__.task = self.pyobject
        return config

    # --- Serialization

    @staticmethod
    def _outputjsonvalue(value, context):
        """Serialize a value"""
        if value is None:
            return None

        elif isinstance(value, (list, tuple)):
            return [ConfigInformation._outputjsonvalue(el, context) for el in value]

        elif isinstance(value, dict):
            return {
                name: ConfigInformation._outputjsonvalue(el, context)
                for name, el in value.items()
            }

        elif isinstance(value, Path):
            return {"type": "path", "value": str(value)}

        elif isinstance(value, SerializedPath):
            return {
                "type": "path.serialized",
                "value": str(value.path),
                "is_folder": value.is_folder,
            }

        elif isinstance(value, (int, float, str)):
            return value

        elif isinstance(value, Enum):
            return {
                "type": "enum",
                "module": value.__class__.__module__,
                "enum": value.__class__.__qualname__,
                "value": value.name,
            }

        elif isinstance(value, Config):
            return {
                "type": "python",
                "value": id(value),
            }

        raise NotImplementedError("Cannot serialize objects of type %s", type(value))

    def __get_objects__(
        self,
        objects: List[Dict],
        context: SerializationContext,
    ) -> List[Dict]:
        """Returns the list of objects necessary to deserialize ourself

        :param objects: the already output objects
        :param context: The serialization context (e.g. useful to create files)
        """

        # Skip if already serialized
        if id(self.pyobject) in context.serialized:
            return objects

        context.serialized.add(id(self.pyobject))

        # Serialize sub-objects
        for argument, value in self.xpmvalues():
            if value is not None:
                ConfigInformation.__collect_objects__(value, objects, context)

        # Adds task
        if self.task is not None and self.task is not self:
            ConfigInformation.__collect_objects__(self.task, objects, context)

        # Serialize initialization tasks
        ConfigInformation.__collect_objects__(self.init_tasks, objects, context)

        # Serialize ourselves
        state_dict = {
            "id": id(self.pyobject),
            "module": self.xpmtype._module,
            "type": self.xpmtype.value_type.__qualname__,
            "typename": self.xpmtype.name(),
            "identifier": self.identifier.state_dict(),
        }

        # Add pre/init tasks
        if self.init_tasks:
            state_dict["init-tasks"] = [id(init_task) for init_task in self.init_tasks]

        if self.meta:
            state_dict["meta"] = self.meta

        if self.task is not None:
            state_dict["task"] = id(self.task)

        if not self.xpmtype._package:
            state_dict["file"] = str(self.xpmtype._file)

        # Serialize identifier and typename
        jsonfields = state_dict["fields"] = {}
        for argument, value in self.xpmvalues():
            with context.push(argument.name) as var_path:
                if argument.is_data and value is not None:
                    assert isinstance(
                        value, Path
                    ), f"Data arguments should be paths (type is {type(value)})"
                    value = context.serialize(var_path, value)

                jsonfields[argument.name] = ConfigInformation._outputjsonvalue(
                    value, context
                )

        objects.append(state_dict)
        return objects

    @staticmethod
    def __collect_objects__(value, objects: List[Dict], context: SerializationContext):
        """Serialize all needed configuration objects, looking at sub
        configurations if necessary"""
        if value is None:
            return

        if isinstance(value, Config):
            value.__xpm__.__get_objects__(objects, context)
        elif isinstance(value, (list, tuple)):
            for el in value:
                ConfigInformation.__collect_objects__(el, objects, context)
        elif isinstance(value, dict):
            for el in value.values():
                ConfigInformation.__collect_objects__(el, objects, context)
        elif isinstance(value, (Path, int, float, str, Enum)):
            pass
        else:
            raise NotImplementedError(
                "Cannot serialize objects of type %s", type(value)
            )

    def outputjson(self, out: io.TextIOBase, context: SerializationContext):
        """Outputs the json of this object

        The format is an array of objects
        {
            "workspace": FOLDERPATH,
            "version": 2,
            "objects": [
                {
                    "id": <ID of the object>,
                    "filename": <filename>, // if in a file
                    "module": <module>, // if in a module
                    "type": <type>, // the type within the module or file
                    "fields":
                        { "key":  {"type": <type>, "value": <value>} }
                }
            ]

        }

        <type> is either a base type or a "python"

        The last object is the one that is serialized

        Note: Tags are no longer stored in params.json. They are managed by the
        experiment state provider (scoped to job_id, experiment_id, run_id) and
        also stored in experiment state.json for full experiment details.

        Arguments:
            out {io.TextIOBase} -- The output stream
            context {[type]} -- the command context
        """
        json.dump(
            {
                "workspace": str(context.workspace.path.absolute()),
                "version": 2,
                "experimaestro": experimaestro.__version__,
                "objects": self.__get_objects__([], context),
            },
            out,
        )

    def __json__(self) -> str:
        """Returns the JSON representation of the object itself"""
        return json.dumps(self.__get_objects__([], SerializationContext()))

    def serialize(self, save_directory: Path):
        """Serialize the configuration and its data files into a directory"""
        context = SerializationContext(save_directory=save_directory)

        with (save_directory / "definition.json").open("wt") as out:
            objects = self.__get_objects__([], context)
            json.dump(objects, out)

    @staticmethod
    def deserialize(
        path: Union[str, Path, SerializedPathLoader],
        as_instance: bool = False,
        return_tasks: bool = False,
    ) -> "Config":
        """Deserialize a configuration

        :param path: The filesystem Path to use, or a way to download the
            information through a function taking two arguments
        :param as_instance: Return an instance
        :return: a Config object, its instance or a tuple (instance, init_tasks) is return_tasks is True
        """
        # Load
        assert not (
            as_instance and return_tasks
        ), "Cannot set as_instance and return_tasks to True"
        if callable(path):
            data_loader = path
        else:
            path = Path(path)

            def data_loader(s: Union[Path, str, SerializedPath]):
                if isinstance(s, SerializedPath):
                    return path / Path(s.path)
                return path / Path(s)

        with data_loader("definition.json").open("rt") as fh:
            config = json.load(fh)

        return ConfigInformation.fromParameters(
            config,
            as_instance=as_instance,
            data_loader=data_loader,
            return_tasks=return_tasks,
        )

    @staticmethod
    def _objectFromParameters(value: Any, objects: Dict[str, Any]):
        # A list
        if isinstance(value, list):
            return [ConfigInformation._objectFromParameters(x, objects) for x in value]

        # A dictionary
        if isinstance(value, dict):
            if "type" not in value:
                # Just a plain dictionary
                return {
                    ConfigInformation._objectFromParameters(
                        key, objects
                    ): ConfigInformation._objectFromParameters(value, objects)
                    for key, value in value.items()
                }

            # The value is an object (that has been serialized first)
            if value["type"] == "python":
                return objects[value["value"]]

            # A path
            if value["type"] == "path":
                return Path(value["value"])

            if value["type"] == "path.serialized":
                return SerializedPath(value["value"], value["is_folder"])

            if value["type"] == "enum":
                module = importlib.import_module(value["module"])
                enumClass = getqualattr(module, value["enum"])
                return enumClass[value["value"]]

            raise Exception("Unhandled type: %s", value["type"])

        # Just a simple value
        return value

    @overload
    @staticmethod
    def fromParameters(  # noqa: E704
        definitions: List[Dict],
        as_instance=True,
        save_directory: Optional[Path] = None,
        discard_id: bool = False,
    ) -> "ConfigMixin": ...

    @overload
    @staticmethod
    def fromParameters(  # noqa: E704
        definitions: List[Dict],
        as_instance=False,
        return_tasks=True,
        save_directory: Optional[Path] = None,
        discard_id: bool = False,
    ) -> Tuple["Config", List["LightweightTask"]]: ...

    @overload
    @staticmethod
    def fromParameters(  # noqa: E704
        definitions: List[Dict],
        as_instance=False,
        save_directory: Optional[Path] = None,
        discard_id: bool = False,
    ) -> "Config": ...

    @staticmethod
    def load_objects(  # noqa: C901
        definitions: List[Dict],
        as_instance=True,
        data_loader: Optional[SerializedPathLoader] = None,
        discard_id: bool = False,
    ):
        """Load the objects"""
        o = None
        objects = {}
        import experimaestro.taskglobals as taskglobals
        from ..identifier import Identifier

        # Loop over all the definitions and create objects
        for definition in definitions:
            module_name = definition["module"]

            # Avoids problem when runing module
            if module_name == "__main__":
                module_name = "_main_"

            if "file" in definition:
                path = definition["file"]
                with add_to_path(str(Path(path).parent)):
                    spec = importlib.util.spec_from_file_location(module_name, path)
                    mod = importlib.util.module_from_spec(spec)
                    sys.modules[module_name] = mod
                    spec.loader.exec_module(mod)
            else:
                try:
                    logger.debug("Importing module %s", definition["module"])
                    mod = importlib.import_module(module_name)
                except ModuleNotFoundError:
                    # More hints on the nature of the error
                    logging.warning(
                        "(1) Either the python path is wrong – %s", ":".join(sys.path)
                    )
                    logging.warning("(2) There is not __init__.py in your module")
                    raise

            cls = getqualattr(mod, definition["type"])

            # Creates an object (or a config)
            if as_instance:
                o = cls.__new__(cls)
            else:
                o = cls.XPMConfig.__new__(cls.XPMConfig)
            assert definition["id"] not in objects, "Duplicate id %s" % definition["id"]
            objects[definition["id"]] = o

        # Now that objects have been created, fill in the fields
        for definition in definitions:
            o = objects[definition["id"]]
            xpmtype = o.__getxpmtype__()  # type: ObjectType

            # If instance...
            if as_instance:
                # ... calls the parameter-less initialization
                o.__init__()

                # ... sets potentially useful properties
                if "typename" in definition:
                    o.__xpmtypename__ = definition["typename"]
                    if not discard_id:
                        o.__xpmidentifier__ = Identifier.from_state_dict(
                            definition["identifier"]
                        )

                    if "." in o.__xpmtypename__:
                        _, name = o.__xpmtypename__.rsplit(".", 1)
                    else:
                        name = o.__xpmtypename__

                    if taskglobals.Env.instance().wspath is not None:
                        basepath = (
                            taskglobals.Env.instance().wspath
                            / "jobs"
                            / o.__xpmtypename__.lower()
                            / o.__xpmidentifier__.all.hex().lower()
                        )
                        o.__xpm_stdout__ = basepath / f"{name.lower()}.out"
                        o.__xpm_stderr__ = basepath / f"{name.lower()}.err"
            else:
                o.__init__()
                xpminfo = o.__xpm__  # type: ConfigInformation
                xpminfo.loaded = True

                meta = definition.get("meta", None)
                if meta:
                    xpminfo._meta = meta
                if xpminfo.xpmtype.task is not None:
                    xpminfo.job = object()

            # Set the fields
            for name, value in definition["fields"].items():
                v = ConfigInformation._objectFromParameters(value, objects)

                # Transform serialized paths arguments
                argument = xpmtype.arguments[name]
                if argument.is_data and v is not None:
                    if isinstance(v, SerializedPath):
                        if data_loader is None:
                            v = v.path
                        else:
                            v = data_loader(v)
                    else:
                        assert isinstance(v, Path), "Excepted Path, got {type(v)}"

                if as_instance:
                    # Unwrap the value if needed
                    setattr(o, name, v)

                    assert (
                        getattr(o, name) is v
                    ), f"Problem with deserialization {name} of {o.__class__}"
                else:
                    o.__xpm__.set(name, v, bypass=True)

            if as_instance:
                # Calls post-init
                o.__post_init__()

            else:
                if task_id := definition.get("task", None):
                    o.__xpm__.task = objects[task_id]

                # Seal and set the identifier
                if not discard_id:
                    xpminfo._identifier = Identifier.from_state_dict(
                        definition["identifier"]
                    )
                xpminfo._sealed = True

        return objects

    @staticmethod
    def fromParameters(  # noqa: C901
        definitions: List[Dict],
        as_instance=True,
        data_loader: Optional[SerializedPathLoader] = None,
        discard_id: bool = False,
        return_tasks: bool = False,
    ):
        # Get the objects
        objects = ConfigInformation.load_objects(
            definitions,
            as_instance=as_instance,
            data_loader=data_loader,
            discard_id=discard_id,
        )

        # Get the last one
        o = objects[definitions[-1]["id"]]

        # Run pre-task (or returns them)
        if as_instance or return_tasks:
            # Collect init tasks
            init_tasks = []
            for init_task_id in definitions[-1].get("init-tasks", []):
                init_task = objects[init_task_id]
                init_tasks.append(init_task)

            if as_instance:
                for init_task in init_tasks:
                    logger.info("Executing init task %s", type(init_task))
                    init_task.execute()
            else:
                return o, init_tasks

        return o

    class FromPython(ConfigWalk):
        def __init__(self, context: ConfigWalkContext, *, objects: ObjectStore = None):
            super().__init__(context)
            self.objects = ObjectStore() if objects is None else objects

        def preprocess(self, config: "Config"):
            if self.objects.is_constructed(id(config)):
                return False, self.objects.retrieve(id(config))
            return True, None

        def stub(self, config: "Config"):
            o = self.objects.retrieve(id(config))

            if o is None:
                # Creates an object (and not a config)
                o = config.__xpmtype__.value_type()

                # Store in cache
                self.objects.add_stub(id(config), o)

            return o

        def postprocess(self, stub, config: "Config", values: Dict[str, Any]):
            # Copy values from the
            for key, value in values.items():
                setattr(stub, key, value)

            # Call __post_init__
            stub.__post_init__()

            self.objects.set_constructed(id(config))
            return stub

    def fromConfig(self, context: ConfigWalkContext, *, objects: ObjectStore = None):
        """Generate an instance given the current configuration"""

        # Validate and seal
        self.validate()
        self.seal(context)

        processor = ConfigInformation.FromPython(context, objects=objects)
        last_object = processor(self.pyobject)

        return last_object

    def add_dependencies(self, *dependencies):
        self.dependencies.extend(dependencies)


def clone(v):
    """Clone a value"""
    if v is None:
        return v

    if isinstance(v, (str, float, int, Path)):
        return v

    if isinstance(v, list):
        return [clone(x) for x in v]

    if isinstance(v, dict):
        return {clone(key): clone(value) for key, value in v.items()}

    if isinstance(v, Enum):
        return v

    if isinstance(v, tuple):
        return tuple(clone(x) for x in v)

    if isinstance(v, Config):
        # Create a new instance
        kwargs = {
            argument.name: clone(value)
            for argument, value in v.__xpm__.xpmvalues()
            if argument.generator is None and not argument.constant
        }

        config = type(v)(**kwargs)
        return config

    raise NotImplementedError("Clone not implemented for type %s" % type(v))


class ConfigMixin:
    """Class for configuration objects"""

    __xpmtype__: ObjectType
    """The associated XPM type"""

    __xpm__: ConfigInformation
    """The __xpm__ object contains all instance specific information about a
    configuration/task"""

    # Set when this instance was created via a deprecated config with replace=True
    _deprecated_from: "ConfigMixin | None" = None

    def __new__(cls, **kwargs):
        """Create a new config instance, handling deprecated replacements."""
        xpmtype = cls.__xpmtype__

        # Check if this is a deprecated type with replace=True
        if xpmtype._deprecation is not None and xpmtype._deprecation.replace:
            # Create the deprecated instance normally
            instance = object.__new__(cls)
            # Initialize it
            ConfigMixin.__init__(instance, **kwargs)
            # Convert to the new type
            converted = instance.__convert__()
            # Mark that this came from a deprecated config
            converted._deprecated_from = instance
            return converted

        # Normal creation
        return object.__new__(cls)

    def __getattribute__(self, name: str):
        """Get an attribute, handling XPM arguments specially.

        We use __getattribute__ instead of __getattr__ because default values
        like `b: Param[X] = None` create class attributes that would prevent
        __getattr__ from being called.
        """
        # Get __xpm__ without recursion
        try:
            xpm = object.__getattribute__(self, "__xpm__")
        except AttributeError:
            # During early init, __xpm__ may not exist yet
            return object.__getattribute__(self, name)

        # Check if this is an XPM argument - parameters take precedence
        xpmtype = object.__getattribute__(self, "__xpmtype__")
        if name in xpmtype.arguments:
            return xpm.get(name)

        # Fall back to normal lookup (methods, etc.)
        return object.__getattribute__(self, name)

    def __setattr__(self, name: str, value):
        """Set an attribute, handling XPM arguments specially."""
        # Allow setting internal attributes directly
        if name in ("__xpm__", "_deprecated_from"):
            object.__setattr__(self, name, value)
            return

        # Check if we have __xpm__ yet (might not during early init)
        xpm = self.__dict__.get("__xpm__")
        if xpm is None:
            object.__setattr__(self, name, value)
            return

        # Check if this is an XPM argument
        xpmtype = self.__xpmtype__
        if name in xpmtype.arguments:
            xpm.set(name, value)
            return

        # Check for deprecated replacement warning
        deprecated_from = self.__dict__.get("_deprecated_from")
        if deprecated_from is not None:
            deprecated_xpmtype = deprecated_from.__xpmtype__
            if name in deprecated_xpmtype.arguments:
                logger.warning(
                    f"Attribute '{name}' was in deprecated config "
                    f"{deprecated_xpmtype.identifier} but is not in "
                    f"{xpmtype.identifier}. The value is being discarded."
                )
                return  # Don't set the attribute

        # Normal attribute setting
        object.__setattr__(self, name, value)

    def __init__(self, **kwargs):
        """Initialize the configuration with the given parameters"""

        # Skip if already initialized (can happen with deprecated replace=True)
        if hasattr(self, "__xpm__"):
            return

        # Add configuration
        xpmtype = self.__xpmtype__

        # Create the config information object
        xpm = ConfigInformation(self)

        # Get the line where the object was created (error reporting)
        caller = inspect.getframeinfo(inspect.stack()[1][0])
        xpm._initinfo = "%s:%s" % (str(Path(caller.filename).absolute()), caller.lineno)

        self.__xpm__ = xpm

        # Initialize with default arguments (or None)
        for name, value in xpmtype.arguments.items():
            if name not in kwargs:
                if value.default is not None:
                    self.__xpm__.set(name, clone(value.default), bypass=True)
                elif not value.required:
                    self.__xpm__.set(name, None, bypass=True)

        # Initialize with arguments
        for name, value in kwargs.items():
            # Check if argument is OK
            if name not in xpmtype.arguments:
                attribute = getattr(self.__class__, name, None)
                if isinstance(attribute, DeprecatedAttribute):
                    attribute.__set__(self, value)
                    continue
                raise ValueError("%s is not an argument for %s" % (name, xpmtype))

            # Special case of a tagged value
            if isinstance(value, TaggedValue):
                value = value.value
                self.__xpm__._tags[name] = value

            # Really set the value
            xpm.set(name, value)

    def __repr__(self):
        return f"Config[{self.__xpmtype__.identifier}]"

    def __str__(self):
        params = ", ".join(
            [f"{key}={value}" for key, value in self.__xpm__.values.items()]
        )
        return (
            f"{self.__xpmtype__.value_type.__module__}."
            f"{self.__xpmtype__.value_type.__qualname__}({params})"
        )

    def tag(self, name, value):
        self.__xpm__.addtag(name, value)
        return self

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        for argument, value in self.__xpm__.xpmvalues():
            if value != getattr(other, argument.name, None):
                return False
        return True

    def __arguments__(self):
        """Returns a map containing all the arguments"""
        return {arg.name: value for arg, value in self.__xpm__.xpmvalues()}

    def tags(self):
        """Returns the tag associated with this object (and below)"""
        return self.__xpm__.tags()

    def add_dependencies(self, *dependencies):
        """Adds tokens to the task"""
        self.__xpm__.add_dependencies(*dependencies)
        return self

    def instance(
        self,
        context: ConfigWalkContext = None,
        *,
        objects: ObjectStore = None,
        keep: bool = True,
    ) -> T:
        """Return an instance with the current values

        :param context: The context when computing the instance
        :param objects: The previously built objects (so that we avoid
            re-creating instances of past configurations)
        :param keep: register a configuration in the __config__ field of the
            instance
        """
        if context is None:
            from experimaestro.xpmutils import EmptyContext

            context = EmptyContext()
        else:
            assert isinstance(
                context, ConfigWalkContext
            ), f"{context.__class__} is not an instance of ConfigWalkContext"

        instance = self.__xpm__.fromConfig(context, objects=objects)  # type: ignore
        if keep:
            object.__setattr__(instance, "__config__", self)
        return instance

    def submit(
        self,
        *,
        workspace=None,
        launcher=None,
        run_mode: "RunMode" = None,
        init_tasks: List["LightweightTask"] = [],
        max_retries: Optional[int] = None,
    ):
        """Submit this task

        :param workspace: the workspace, defaults to None
        :param launcher: The launcher, defaults to None
        :param run_mode: Run mode (if None, uses the workspace default)
        :param max_retries: Maximum number of retries for resumable tasks that timeout (default: from workspace settings or 3)
        :return: an object object
        """
        return self.__xpm__.submit(
            workspace,
            launcher,
            run_mode=run_mode,
            init_tasks=init_tasks,
            max_retries=max_retries,
        )

    def stdout(self):
        return self.__xpm__.job.stdout

    def stderr(self):
        return self.__xpm__.job.stderr

    def wait(self):
        return self.__xpm__.job.wait()

    @property
    def job(self):
        return self.__xpm__.job

    @property
    def jobpath(self):
        if self.__xpm__.job:
            return self.__xpm__.job.jobpath
        raise AssertionError("Cannot ask the job path of a non submitted task")

    def copy(self):
        """Returns a copy of this configuration (ignores other non parameters
        attributes)"""
        return clone(self)

    def copy_dependencies(self, other: "ConfigMixin"):
        """Add all the dependencies from other configuration"""

        # Add task dependency
        if other.__xpm__.task is not None:
            assert self.__xpm__.task is None
            self.__xpm__.task = other.__xpm__.task

        # Add other dependencies
        self.__xpm__.add_dependencies(*other.__xpm__.dependencies)

    def __rmatmul__(self, other: "ConfigMixin") -> "ConfigMixin":
        """Right-associative composition operator: B() @ A(x=1) is equivalent to B(a=A(x=1))

        For expression `other @ self`, finds the unique parameter in `other` that
        accepts `self`'s type and sets it. Returns `self` to enable right-associative
        chaining: `Outer() @ Middle() @ Inner()` builds Outer(middle=Middle(inner=Inner())).

        The chain is evaluated left-to-right by Python, but returns the inner config
        so each step adds the current result as a parameter to the next outer config.

        :param other: The outer configuration that will receive self as a parameter
        :return: self (the inner configuration) to continue the chain
        """
        if not isinstance(other, ConfigMixin):
            return NotImplemented

        # Find parameters in 'other' that can accept self's type
        self_type = self.__xpmtype__.value_type
        matching_params = []

        for name, argument in other.__xpmtype__.arguments.items():
            # Get the expected type for this argument
            arg_type = argument.type
            if hasattr(arg_type, "value_type"):
                # It's an ObjectType wrapper
                expected_type = arg_type.value_type
            elif hasattr(arg_type, "__origin__"):
                # Generic type like Optional[X] or List[X]
                continue  # Skip complex types for now
            elif isinstance(arg_type, type):
                expected_type = arg_type
            else:
                continue

            # Check if self's type is compatible
            if isinstance(expected_type, type) and issubclass(self_type, expected_type):
                matching_params.append(name)

        if len(matching_params) == 0:
            raise ValueError(
                f"No parameter in {other.__xpmtype__} accepts type {self_type.__name__}"
            )
        if len(matching_params) > 1:
            raise ValueError(
                f"Ambiguous composition: parameters {matching_params} in "
                f"{other.__xpmtype__} all accept type {self_type.__name__}"
            )

        # Set the parameter on 'other'
        param_name = matching_params[0]
        other.__xpm__.set(param_name, self)
        return other


class Config:
    """Base type for all objects in python interface"""

    __xpmid__: ClassVar[Optional[str]]
    """Optional configuration ID, mostly useful when moving a class to another
    package to avoid changes in computed task identifiers"""

    __xpmtype__: ClassVar[ObjectType]
    """The object type holds all the information about a specific subclass
    experimaestro metadata"""

    @classproperty
    def XPMConfig(cls):
        if issubclass(cls, ConfigMixin):
            return cls
        return cls.__getxpmtype__().config_type

    @classproperty
    def C(cls):
        """Alias for XPMConfig"""
        return cls.XPMConfig

    @classproperty
    def XPMValue(cls):
        """Get the value class for this configuration.

        Returns the explicitly registered value class, or the base config class
        if no value class was registered.
        """
        return cls.__getxpmtype__().value_type

    @classmethod
    def value_class(cls):
        """Decorator to register an external value class for this configuration.

        This allows declaring a separate class that will be used when creating
        instances, which is useful to avoid initializing resources (e.g., PyTorch)
        when only configuring.

        Example:
            class Model(Config):
                hidden_size: Param[int]

            @Model.value_class()
            class TorchModel(Model, nn.Module):
                def __init__(self):
                    super().__init__()
                    self.layer = nn.Linear(self.hidden_size, self.hidden_size)

        The value class must:
        1. Be a subclass of the configuration class
        2. Be a subclass of parent configuration value classes (if any)
        """

        def decorator(value_class: type) -> type:
            xpmtype = cls.__getxpmtype__()

            # Check that value class is a subclass of the config class
            if not issubclass(value_class, cls):
                raise TypeError(
                    f"Value class {value_class.__name__} must be a subclass of "
                    f"{cls.__name__}"
                )

            # Check that value class inherits from parent value classes
            for base in cls.__bases__:
                if base is Config or not issubclass(base, Config):
                    continue
                parent_xpmtype = base.__getxpmtype__()
                # Check if parent has an explicit value type (different from original)
                if parent_xpmtype.value_type is not parent_xpmtype._original_type:
                    parent_value = parent_xpmtype.value_type
                    if not issubclass(value_class, parent_value):
                        raise TypeError(
                            f"Value class {value_class.__name__} must be a subclass of "
                            f"parent value class {parent_value.__name__}"
                        )

            # Register the value class
            xpmtype.set_value_type(value_class)
            return value_class

        return decorator

    @classmethod
    def __getxpmtype__(cls) -> "ObjectType":
        """Get (and create if necessary) the Object type associated
        with thie Config object"""
        xpmtype = cls.__dict__.get("__xpmtype__", None)
        if xpmtype is None:
            from experimaestro.core.types import ObjectType

            # Will set __xpmtype__
            try:
                return ObjectType(cls)
            except Exception:
                logger.error("Error while creating object type for %s", cls)
                raise
        return xpmtype

    def __validate__(self):
        """Validate the values"""
        pass

    def __post_init__(self):
        """Called after the object  __init__() and with properties set"""
        # Default implementation is to do nothing
        pass

    def __json__(self):
        """Returns a JSON version of the object (if possible)"""
        return self.__xpm__.__json__()

    def __identifier__(self) -> "Identifier":
        return self.__xpm__.identifier

    def copy_dependencies(self, other: "Config"):
        """Add pre-tasks from the listed configurations"""
        raise AssertionError(
            "The 'copy_dependencies' method can only be used during configuration"
        )

    def register_task_output(self, method, *args, **kwargs):
        # Determine the path for this...
        path = taskglobals.Env.instance().xpm_path / "task-outputs.jsonl"
        path.parent.mkdir(parents=True, exist_ok=True)

        data = json.dumps(
            {
                "key": f"{self.__xpmidentifier__}/{method.__name__}",
                "args": args,
                "kwargs": kwargs,
            }
        )
        with path.open("at") as fp:
            fp.writelines([data, "\n"])
            fp.flush()


class InstanceConfig(Config):
    """Base class for configurations where instance identity matters.

    When a Config class derives from InstanceConfig instead of Config,
    instances are distinguished based on their object identity when used
    in containers. This enables distinguishing between shared and separate
    instances even when all parameters are identical.

    Example:
        >>> class SubModel(InstanceConfig):
        ...     value: Param[int] = 100
        >>> class MainModel(Config):
        ...     m1: Param[SubModel]
        ...     m2: Param[SubModel]
        >>>
        >>> sm1 = SubModel.C()
        >>> sm2 = SubModel.C()  # Same params, different instance
        >>>
        >>> # Shared instance (same object used twice)
        >>> shared = MainModel.C(m1=sm1, m2=sm1)
        >>>
        >>> # Separate instances (different objects)
        >>> separate = MainModel.C(m1=sm1, m2=sm2)
        >>>
        >>> # Different identifiers: shared vs separate
        >>> shared.__identifier__() != separate.__identifier__()

    The instance order is determined by the traversal order during
    identifier computation, ensuring reproducibility.
    """

    pass


class LightweightTask(Config):
    """A task that can be run before or after a real task to modify its behaviour"""

    def execute(self):
        raise NotImplementedError()


class Task(LightweightTask):
    """Base class for tasks"""

    __tags__: Dict[str, str]
    """Tags associated with class"""

    def submit(self):
        raise AssertionError("This method can only be used during configuration")

    def watch_output(self, method, callback):
        """Sets up a callback

        :param method: a method within a configuration
        :param callback: the callback
        """
        self.__xpm__.watch_output(method, callback)

    def on_completed(self, callback: Callable[[], None]):
        self.__xpm__.on_completed(callback)


# --- Utility functions


def copyconfig(config: Config, **kwargs):
    """Copy a configuration

    Useful to modify a configuration that can be potentially
    wrapped into a task output (i.e., the configuration can be
    a task output). If the configuration is sealed, the copy
    will be unsealed.

    :param config: _description_
    :param kwargs: Modify the configuration by assigning the values

    :return: _description_
    """

    # Builds a new configuration object
    copy = config.__class__()

    fullkwargs = {name: value for name, value in config.__xpm__.values.items()}
    fullkwargs.update(kwargs)
    for name, value in fullkwargs.items():
        copy.__xpm__.set(name, value, True)

    # Remove generated attributes
    for argument, value in copy.__xpm__.xpmvalues(generated=True):
        if argument.generator is not None and value is not None:
            del copy.__xpm__.values[argument.name]

    return copy


def setmeta(config: Config, flag: bool):
    """Flags the configuration as a meta-parameter"""
    config.__xpm__.set_meta(flag)
    return config


class ResumableTask(Task):
    """Base class for resumable/checkpointable tasks

    Resumable tasks can be restarted if they are stopped by a time limit
    (e.g., SLURM job timeout). The task directory and dynamic outputs are
    preserved across restarts to allow checkpoint recovery.
    """

    def remaining_time(self) -> Optional[float]:
        """Returns the remaining time in seconds before the job times out.

        This is useful for checkpointing before hitting a time limit
        (e.g., SLURM walltime).

        Returns:
            The remaining time in seconds, or None if:
            - There is no time limit
            - The launcher doesn't support querying remaining time
            - The task is not running
        """
        launcher_info = taskglobals.Env.instance().launcher_info
        if launcher_info is None:
            return None
        return launcher_info.remaining_time()


def cache(fn, name: str):
    def __call__(config, *args, **kwargs):
        import experimaestro.taskglobals as taskglobals

        # Get path and create directory if needed
        hexid = config.__xpmidentifier__  # type: Identifier
        typename = config.__xpmtypename__  # type: str
        dir = taskglobals.Env.instance().wspath / "config" / typename / hexid.all.hex()

        if not dir.exists():
            dir.mkdir(parents=True, exist_ok=True)

        path = dir / name
        ipc_lock = fasteners.InterProcessLock(path.with_suffix(path.suffix + ".lock"))
        with ipc_lock:
            r = fn(config, path, *args, **kwargs)
            return r

    return __call__



from collections import ChainMap
from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import Optional
from experimaestro.settings import WorkspaceSettings, Settings


# Current workspace version
WORKSPACE_VERSION = 0


class RunMode(str, Enum):
    NORMAL = "normal"
    """Normal run"""

    GENERATE_ONLY = "generate"
    """Do not run, but generate the params.json file"""

    DRY_RUN = "dry-run"
    """Do not run"""


class Workspace:
    """Workspace environment for experiments

    This is a simple container for workspace settings, environment, and configuration.
    Multiple Workspace instances can exist for the same path - the singleton pattern
    is handled by WorkspaceStateProvider which manages the database per workspace path.
    """

    CURRENT = None
    settings: "Settings"
    workspace_settings: "WorkspaceSettings"

    def __init__(
        self,
        settings: "Settings",
        workspace_settings: "WorkspaceSettings",
        launcher=None,
        run_mode: RunMode = None,
    ):
        """Initialize workspace environment

        Args:
            settings: Global settings
            workspace_settings: Workspace-specific settings
            launcher: Default launcher for this workspace
            run_mode: Run mode for experiments in this workspace
        """
        self.settings = settings
        self.workspace_settings = workspace_settings

        path = self.workspace_settings.path
        self.notificationURL: Optional[str] = None
        if isinstance(path, Path):
            path = path.absolute()
        else:
            path = Path(path).absolute()
        self.path = path
        self.run_mode = run_mode
        self.python_path = []

        from ..launchers import Launcher

        self.launcher = launcher or Launcher.get(path)

        self.env = ChainMap({}, workspace_settings.env, settings.env)

        # Reference counting for nested context managers
        self._ref_count = 0

    def __enter__(self):
        # Increment reference count
        self._ref_count += 1

        # Only initialize on first entry
        if self._ref_count == 1:
            # Check if a different workspace is already active
            if Workspace.CURRENT is not None and Workspace.CURRENT.path != self.path:
                raise RuntimeError(
                    f"Cannot activate workspace at {self.path} - "
                    f"workspace at {Workspace.CURRENT.path} is already active. "
                    "Multiple workspaces are not yet supported."
                )

            self.old_workspace = Workspace.CURRENT
            Workspace.CURRENT = self

        return self

    def __exit__(self, *args):
        # Decrement reference count
        self._ref_count -= 1

        # Only cleanup on last exit
        if self._ref_count == 0:
            Workspace.CURRENT = self.old_workspace

    @cached_property
    def alt_workspaces(self):
        for ws_id in self.workspace_settings.alt_workspaces:
            yield self.settings.workspaces[ws_id]

    @property
    def alt_workdirs(self):
        yield from map(lambda ws: ws.path, self.workspace_settings.alt_workspaces)

    @property
    def connector(self):
        """Returns the default connector"""
        return self.launcher.connector

    @property
    def jobspath(self):
        """Folder for jobs"""
        return self.path / "jobs"

    @property
    def experimentspath(self):
        """Folder for experiments"""
        return self.path / "xp"

    @property
    def configcachepath(self):
        """Folder for jobs"""
        return self.path / "config"



from contextlib import contextmanager
from enum import Enum
from functools import cached_property
import hashlib
import logging
import os
import struct
from typing import Optional
from experimaestro.core.objects import Config, ConfigMixin


class ConfigPath:
    """Used to keep track of cycles when computing a hash"""

    def __init__(self):
        self.loops: list[bool] = []
        """Indicates whether a loop was detected up to this node"""

        self.config2index = {}
        """Associate an index in the list with a configuration"""

        self.instance_order: dict[bytes, list[int]] = {}
        """Maps InstanceConfig base identifier to list of object ids in encounter order"""

    def detect_loop(self, config) -> Optional[int]:
        """If there is a loop, return the relative index and update the path"""
        index = self.config2index.get(id(config), None)
        if index is not None:
            for i in range(index, self.depth):
                self.loops[i] = True
            return self.depth - index
        return None

    def has_loop(self):
        return self.loops[-1]

    @property
    def depth(self):
        return len(self.loops)

    @contextmanager
    def push(self, config):
        config_id = id(config)
        assert config_id not in self.config2index

        self.config2index[config_id] = self.depth
        self.loops.append(False)

        try:
            yield
        finally:
            self.loops.pop()
            del self.config2index[config_id]


hash_logger = logging.getLogger("xpm.hash")


def is_ignored(value):
    """Returns True if the value should be ignored by itself"""
    return value is not None and isinstance(value, Config) and (value.__xpm__.meta)


def remove_meta(value):
    """Cleanup a dict/list by removing ignored values"""
    if isinstance(value, list):
        return [el for el in value if not is_ignored(el)]
    if isinstance(value, dict):
        return {key: value for key, value in value.items() if not is_ignored(value)}
    return value


class Identifier:
    def __init__(self, main: bytes):
        self.main = main
        self.has_loops = False

    @cached_property
    def all(self):
        """Returns the overall identifier"""
        return self.main

    def __hash__(self) -> int:
        return hash(self.main)

    def state_dict(self):
        return self.main.hex()

    def __eq__(self, other: object):
        if not isinstance(other, Identifier):
            return False
        return self.main == other.main

    @staticmethod
    def from_state_dict(data: dict[str, str] | str):
        if isinstance(data, str):
            return Identifier(bytes.fromhex(data))

        return Identifier(bytes.fromhex(data["main"]))

    def __repr__(self):
        return self.main.hex()


class IdentifierComputer:
    """This class is in charge of computing a config/task identifier"""

    OBJECT_ID = b"\x00"
    INT_ID = b"\x01"
    FLOAT_ID = b"\x02"
    STR_ID = b"\x03"
    PATH_ID = b"\x04"
    NAME_ID = b"\x05"
    NONE_ID = b"\x06"
    LIST_ID = b"\x07"
    TASK_ID = b"\x08"
    DICT_ID = b"\x09"
    ENUM_ID = b"\x0a"
    CYCLE_REFERENCE = b"\x0b"
    INIT_TASKS = b"\x0c"
    INSTANCE_ID = b"\x0d"

    def __init__(self, config: "ConfigMixin", config_path: ConfigPath, *, version=None):
        # Hasher for parameters
        self._hasher = hashlib.sha256()
        self.config = config
        self.config_path = config_path
        self.version = version or int(os.environ.get("XPM_HASH_COMPUTER", 2))
        if hash_logger.isEnabledFor(logging.DEBUG):
            hash_logger.debug(
                "starting hash (%s): %s", hash(str(self.config)), self.config
            )

    def identifier(self) -> Identifier:
        main = self._hasher.digest()
        if hash_logger.isEnabledFor(logging.DEBUG):
            hash_logger.debug("hash (%s): %s", hash(str(self.config)), str(main))
        return Identifier(main)

    def _hashupdate(self, bytes: bytes):
        """Update the hash computers with some bytes"""
        if hash_logger.isEnabledFor(logging.DEBUG):
            hash_logger.debug(
                "updating hash (%s): %s", hash(str(self.config)), str(bytes)
            )
        self._hasher.update(bytes)

    def update(self, value, *, myself=False):  # noqa: C901
        """Update the hash

        :param value: The value to add to the hash
        :param myself: True if the value is the configuration for which we wish
            to compute the identifier, defaults to False
        :raises NotImplementedError: If the value cannot be processed
        """
        if value is None:
            self._hashupdate(IdentifierComputer.NONE_ID)
        elif isinstance(value, float):
            self._hashupdate(IdentifierComputer.FLOAT_ID)
            self._hashupdate(struct.pack("!d", value))
        elif isinstance(value, int):
            self._hashupdate(IdentifierComputer.INT_ID)
            self._hashupdate(struct.pack("!q", value))
        elif isinstance(value, str):
            self._hashupdate(IdentifierComputer.STR_ID)
            self._hashupdate(value.encode("utf-8"))
        elif isinstance(value, list):
            values = [el for el in value if not is_ignored(el)]
            self._hashupdate(IdentifierComputer.LIST_ID)
            self._hashupdate(struct.pack("!d", len(values)))
            for x in values:
                self.update(x)
        elif isinstance(value, Enum):
            self._hashupdate(IdentifierComputer.ENUM_ID)
            k = value.__class__
            self._hashupdate(
                f"{k.__module__}.{k.__qualname__}:{value.name}".encode("utf-8"),
            )
        elif isinstance(value, dict):
            self._hashupdate(IdentifierComputer.DICT_ID)
            items = [
                (key, value) for key, value in value.items() if not is_ignored(value)
            ]
            items.sort(key=lambda x: x[0])
            for key, value in items:
                self.update(key)
                self.update(value)

        # Handles configurations
        elif isinstance(value, ConfigMixin):
            # For deprecated configs with __convert__, compute identifier from converted config
            # This must happen BEFORE hashing OBJECT_ID to ensure identical identifiers
            if myself:
                xpmtype = value.__xpmtype__
                if xpmtype.deprecated and hasattr(value, "__convert__"):
                    hash_logger.debug(
                        "Computing hash via __convert__ for deprecated %s", xpmtype
                    )
                    converted = value.__convert__()
                    # Delegate entirely to the converted config
                    self.update(converted, myself=True)
                    return

            # Encodes the identifier
            self._hashupdate(IdentifierComputer.OBJECT_ID)

            # If we encode another config, then
            if not myself:
                if loop_ix := self.config_path.detect_loop(value):
                    # Loop detected: use cycle reference
                    self._hashupdate(IdentifierComputer.CYCLE_REFERENCE)
                    self._hashupdate(struct.pack("!q", loop_ix))

                else:
                    # Just use the object identifier
                    value_id = IdentifierComputer.compute(
                        value, version=self.version, config_path=self.config_path
                    )
                    self._hashupdate(value_id.all)

                    # Handle InstanceConfig: add instance order only if duplicate detected
                    from experimaestro.core.objects import InstanceConfig

                    if isinstance(value, InstanceConfig):
                        # Use the base identifier to track instances
                        base_id = value_id.all
                        obj_id = id(value)

                        # Track instances with this base identifier
                        if base_id not in self.config_path.instance_order:
                            # First occurrence - just record it (NO-OP for backwards compatibility)
                            self.config_path.instance_order[base_id] = [obj_id]
                        else:
                            # Check if this specific object was already seen
                            instances = self.config_path.instance_order[base_id]
                            if obj_id not in instances:
                                # New instance with same params - add to list
                                instances.append(obj_id)

                            # Get position in the list
                            position = instances.index(obj_id)

                            # Only add instance order if not the first (position > 0)
                            if position > 0:
                                self._hashupdate(IdentifierComputer.INSTANCE_ID)
                                self._hashupdate(struct.pack("!q", position))

                # And that's it!
                return

            # Process tasks
            if value.__xpm__.task is not None and (value.__xpm__.task is not value):
                hash_logger.debug("Computing hash for task %s", value.__xpm__.task)
                self._hashupdate(IdentifierComputer.TASK_ID)
                self.update(value.__xpm__.task)

            xpmtype = value.__xpmtype__
            self._hashupdate(xpmtype.identifier.name.encode("utf-8"))

            # Process arguments (sort by name to ensure uniqueness)
            arguments = sorted(xpmtype.arguments.values(), key=lambda a: a.name)
            for argument in arguments:
                # Ignored argument
                if argument.ignored:
                    argvalue = value.__xpm__.values.get(argument.name, None)

                    # ... unless meta is set to false
                    if (
                        argvalue is None
                        or not isinstance(argvalue, Config)
                        or (argvalue.__xpm__.meta is not False)
                    ):
                        continue

                if argument.generator:
                    continue

                # Argument value
                # Skip if the argument is not a constant, and
                # - optional argument: both value and default are None
                # - the argument value is equal to the default value AND
                #   ignore_default_in_identifier is True
                try:
                    argvalue = getattr(value, argument.name, None)
                except KeyError:
                    logging.warning(
                        "Parameter %s has not been set in %s created at %s",
                        argument.name,
                        value,
                        value.__xpm__._initinfo,
                    )
                    raise
                if not argument.constant and (
                    (
                        not argument.required
                        and argument.default is None
                        and argvalue is None
                    )
                    or (
                        argument.ignore_default_in_identifier
                        and argument.default is not None
                        and argument.default == remove_meta(argvalue)
                    )
                ):
                    # No update if same value (and not constant)
                    continue

                if (
                    argvalue is not None
                    and isinstance(argvalue, Config)
                    and argvalue.__xpm__.meta
                ):
                    continue

                # Hash name
                self.update(argument.name)

                # Hash value
                self._hashupdate(IdentifierComputer.NAME_ID)
                self.update(argvalue)

            # Add init tasks
            if value.__xpm__.init_tasks:
                self._hashupdate(IdentifierComputer.INIT_TASKS)
                for init_task in value.__xpm__.init_tasks:
                    self.update(init_task)
        else:
            raise NotImplementedError("Cannot compute hash of type %s" % type(value))

    @staticmethod
    def compute(
        config: "ConfigMixin", config_path: ConfigPath | None = None, version=None
    ) -> Identifier:
        """Compute the identifier for a configuration

        :param config: the configuration for which we compute the identifier
        :param config_path: used to track down cycles between configurations
        :param version: version for the hash computation (None for the last one)
        """

        # Try to use the cached value first
        # (if there are no loops)
        if config.__xpm__._sealed:
            identifier = config.__xpm__._identifier
            if identifier is not None and not identifier.has_loops:
                return identifier

        config_path = config_path or ConfigPath()

        with config_path.push(config):
            self = IdentifierComputer(config, config_path, version=version)
            self.update(config, myself=True)
            identifier = self.identifier()
            identifier.has_loops = config_path.has_loop()

        return identifier



import os
from omegaconf import OmegaConf, SCMode
from dataclasses import field, dataclass
from functools import lru_cache
from pathlib import Path
from typing import Dict, Optional, List
import logging
import fnmatch


@dataclass
class ServerSettings:
    port: Optional[int] = None
    """Port for the server"""

    host: Optional[str] = None
    """Hostname for the server"""

    autohost: Optional[str] = None
    """Automatic hostname: values can be `fqdn` or `name`"""

    token: Optional[str] = None
    """Token for the server"""


@dataclass
class WorkspaceSettings:
    """Defines the workspace"""

    id: str
    """The workspace identifier"""

    path: Path = field()
    """The workspace path"""

    env: Dict[str, str] = field(default_factory=dict)
    """Workspace specific environment variables"""

    alt_workspaces: List[str] = field(default_factory=list)
    """Alternative workspaces to find jobs or experiments"""

    max_retries: int = 3
    """Maximum number of retries for resumable tasks that timeout (default: 3)"""

    triggers: List[str] = field(default_factory=list)
    """Glob patterns to automatically select this workspace based on experiment ID"""

    def __post_init__(self):
        self.path = self.path.expanduser().resolve()


@dataclass
class Settings:
    server: ServerSettings = field(default_factory=ServerSettings)
    workspaces: List[WorkspaceSettings] = field(default_factory=list)

    env: Dict[str, str] = field(default_factory=dict)
    """Default environment variables"""


@lru_cache()
def get_settings(path: Optional[Path] = None) -> Settings:
    if "PYTEST_CURRENT_TEST" in os.environ:
        return Settings()
    else:
        schema = OmegaConf.structured(Settings)

        path = path or Path("~/.config/experimaestro/settings.yaml").expanduser()
        if not path.is_file():
            return OmegaConf.to_container(
                schema, structured_config_mode=SCMode.INSTANTIATE
            )

        conf = OmegaConf.load(path)
        return OmegaConf.to_container(
            OmegaConf.merge(schema, conf), structured_config_mode=SCMode.INSTANTIATE
        )


def get_workspace(id: Optional[str] = None) -> Optional[WorkspaceSettings]:
    """Return the workspace settings given an id (or None for the default one)"""
    workspaces = get_settings().workspaces
    if workspaces:
        if id is None:
            return workspaces[0]
        for workspace in workspaces:
            if id == workspace.id:
                return workspace

    return None


def find_workspace(
    *,
    workspace: Optional[str] = None,
    workdir: Optional[Path] = None,
    experiment_id: Optional[str] = None,
) -> WorkspaceSettings:
    """Find workspace

    Args:
        workspace: Explicit workspace ID to use
        workdir: Explicit working directory path
        experiment_id: Experiment ID to match against workspace triggers

    Returns:
        WorkspaceSettings object
    """
    workdir = Path(workdir) if workdir else None

    if workspace:
        ws_env = get_workspace(workspace)
        if ws_env is None:
            raise RuntimeError("No workspace named %s", workspace)

        logging.info("Using workspace %s", ws_env.id)
        if workdir:
            # Overrides working directory
            logging.info(" override working directory: %s", workdir)
            ws_env.path = workdir
    elif workdir:
        logging.info("Using workdir %s", workdir)
        ws_env = WorkspaceSettings("", workdir)
    else:
        # Try to match experiment_id against workspace triggers
        matched_workspace = None
        if experiment_id:
            workspaces = get_settings().workspaces
            for ws in workspaces:
                for trigger in ws.triggers:
                    if fnmatch.fnmatch(experiment_id, trigger):
                        matched_workspace = ws
                        logging.info(
                            "Auto-selected workspace %s (matched trigger '%s')",
                            ws.id,
                            trigger,
                        )
                        break
                if matched_workspace:
                    break

        if matched_workspace:
            ws_env = matched_workspace
        else:
            ws_env = get_workspace()
            assert ws_env is not None, "No workdir or workspace defined, and no default"
            logging.info("Using default workspace %s", ws_env.id)

    return ws_env



"""Unit tests for token locking mechanism

Tests the CounterToken condition variable-based synchronization
without requiring full scheduler integration.
"""

import asyncio
import pytest
import tempfile
from pathlib import Path
import time

from experimaestro.tokens import CounterToken
from experimaestro.locking import LockError

pytestmark = pytest.mark.anyio


async def test_token_acquire_release():
    """Test basic token acquire and release"""
    with tempfile.TemporaryDirectory() as tmpdir:
        token = CounterToken("test-basic", Path(tmpdir) / "token", count=1)

        # Create a mock job target
        class MockJob:
            @property
            def identifier(self):
                return "mock-job-1"

            @property
            def basepath(self):
                return Path(tmpdir) / "job1"

        job = MockJob()

        # Create dependency
        dep = token.dependency(1)
        dep.target = job

        # Should be able to acquire
        lock = await dep.aio_lock(timeout=1.0)
        assert lock is not None
        assert token.available == 0

        # Release
        lock.release()
        assert token.available == 1


async def test_token_blocking():
    """Test that acquiring blocks when no tokens available"""
    with tempfile.TemporaryDirectory() as tmpdir:
        token = CounterToken("test-blocking", Path(tmpdir) / "token", count=1)

        class MockJob:
            def __init__(self, name):
                self.name = name

            @property
            def identifier(self):
                return f"mock-job-{self.name}"

            @property
            def basepath(self):
                return Path(tmpdir) / self.name

        job1 = MockJob("1")
        job2 = MockJob("2")

        dep1 = token.dependency(1)
        dep1.target = job1

        dep2 = token.dependency(1)
        dep2.target = job2

        # Acquire with first dependency
        lock1 = await dep1.aio_lock(timeout=0.5)
        assert token.available == 0

        # Second acquire should timeout
        start = time.time()
        with pytest.raises(LockError, match="Timeout"):
            await dep2.aio_lock(timeout=0.5)
        elapsed = time.time() - start
        assert 0.4 < elapsed < 0.7  # Should timeout around 0.5s

        # Release first lock
        lock1.release()
        assert token.available == 1

        # Now second should succeed
        lock2 = await dep2.aio_lock(timeout=0.5)
        assert lock2 is not None
        lock2.release()


async def test_token_notification():
    """Test that condition notification wakes up waiting tasks"""
    with tempfile.TemporaryDirectory() as tmpdir:
        token = CounterToken("test-notify", Path(tmpdir) / "token", count=1)

        class MockJob:
            def __init__(self, name):
                self.name = name

            @property
            def identifier(self):
                return f"mock-job-{self.name}"

            @property
            def basepath(self):
                return Path(tmpdir) / self.name

        job1 = MockJob("1")
        job2 = MockJob("2")

        dep1 = token.dependency(1)
        dep1.target = job1

        dep2 = token.dependency(1)
        dep2.target = job2

        # Acquire with first dependency
        lock1 = await dep1.aio_lock(timeout=0.5)

        # Start second acquisition in background
        async def acquire_second():
            lock = await dep2.aio_lock(timeout=5.0)  # Long timeout
            return lock

        task = asyncio.create_task(acquire_second())

        # Give it time to start waiting
        await asyncio.sleep(0.1)

        # Release first lock - should notify waiting task
        start = time.time()
        lock1.release()

        # Second task should complete quickly (not timeout)
        lock2 = await task
        elapsed = time.time() - start

        assert lock2 is not None
        assert elapsed < 1.0  # Should wake up immediately, not wait 5s
        lock2.release()


async def test_token_multiple_waiting():
    """Test multiple tasks waiting for tokens"""
    with tempfile.TemporaryDirectory() as tmpdir:
        token = CounterToken("test-multiple", Path(tmpdir) / "token", count=1)

        class MockJob:
            def __init__(self, name):
                self.name = name

            @property
            def identifier(self):
                return f"mock-job-{self.name}"

            @property
            def basepath(self):
                return Path(tmpdir) / self.name

        # Acquire the token
        job1 = MockJob("1")
        dep1 = token.dependency(1)
        dep1.target = job1
        lock1 = await dep1.aio_lock(timeout=0.5)

        # Start multiple waiting tasks
        acquired_order = []

        async def acquire_task(name):
            job = MockJob(name)
            dep = token.dependency(1)
            dep.target = job
            lock = await dep.aio_lock(timeout=10.0)
            acquired_order.append(name)
            await asyncio.sleep(0.05)  # Hold briefly
            lock.release()

        tasks = [
            asyncio.create_task(acquire_task("2")),
            asyncio.create_task(acquire_task("3")),
            asyncio.create_task(acquire_task("4")),
        ]

        # Give tasks time to start waiting
        await asyncio.sleep(0.1)

        # Release first lock
        lock1.release()

        # Wait for all tasks to complete
        await asyncio.gather(*tasks)

        # All tasks should have acquired the lock
        assert len(acquired_order) == 3
        assert set(acquired_order) == {"2", "3", "4"}


async def test_token_timeout_zero():
    """Test that timeout=0 waits indefinitely"""
    with tempfile.TemporaryDirectory() as tmpdir:
        token = CounterToken("test-timeout-zero", Path(tmpdir) / "token", count=1)

        class MockJob:
            def __init__(self, name):
                self.name = name

            @property
            def identifier(self):
                return f"mock-job-{self.name}"

            @property
            def basepath(self):
                return Path(tmpdir) / self.name

        job1 = MockJob("1")
        job2 = MockJob("2")

        dep1 = token.dependency(1)
        dep1.target = job1

        dep2 = token.dependency(1)
        dep2.target = job2

        # Acquire with first
        lock1 = await dep1.aio_lock(timeout=0.5)

        # Start waiting with timeout=0 (infinite)
        async def acquire_infinite():
            return await dep2.aio_lock(timeout=0)  # Should wait forever

        task = asyncio.create_task(acquire_infinite())

        # Give it time to start waiting
        await asyncio.sleep(0.1)

        # Wait a bit more - task should still be waiting
        await asyncio.sleep(0.5)
        assert not task.done()

        # Release first lock
        lock1.release()

        # Now task should complete
        lock2 = await asyncio.wait_for(task, timeout=2.0)
        assert lock2 is not None
        lock2.release()



"""Functional tests for CLI jobs commands

Tests the jobs list, kill, and clean commands using the WorkspaceStateProvider.
"""

import pytest
import time
from click.testing import CliRunner

from experimaestro.cli import cli
from experimaestro.scheduler.state_provider import WorkspaceStateProvider
from experimaestro.scheduler.state_db import (
    initialize_workspace_database,
    close_workspace_database,
    ExperimentModel,
    ExperimentRunModel,
    JobModel,
    ALL_MODELS,
)
from experimaestro.scheduler.workspace import WORKSPACE_VERSION


@pytest.fixture
def workspace_path(tmp_path):
    """Create a workspace directory with database initialized"""
    ws_path = tmp_path / "workspace"
    ws_path.mkdir()

    # Create version file with current workspace version
    (ws_path / ".__experimaestro__").write_text(str(WORKSPACE_VERSION))

    # Initialize database
    xpm_dir = ws_path / ".experimaestro"
    xpm_dir.mkdir()
    db_path = xpm_dir / "workspace.db"
    db = initialize_workspace_database(db_path, read_only=False)

    # Create jobs directory
    jobs_dir = ws_path / "jobs"
    jobs_dir.mkdir()

    yield ws_path

    close_workspace_database(db)


@pytest.fixture
def workspace_with_jobs(workspace_path):
    """Create a workspace with some test jobs in the database"""
    db_path = workspace_path / ".experimaestro" / "workspace.db"
    db = initialize_workspace_database(db_path, read_only=False)

    with db.bind_ctx(ALL_MODELS):
        # Create experiment and run
        ExperimentModel.create(experiment_id="test_exp", current_run_id="run_001")
        ExperimentRunModel.create(
            experiment_id="test_exp", run_id="run_001", status="active"
        )

        # Create jobs with different states
        jobs_data = [
            ("job_done_1", "mymodule.DoneTask", "done"),
            ("job_done_2", "mymodule.DoneTask", "done"),
            ("job_error_1", "mymodule.ErrorTask", "error"),
            ("job_running_1", "mymodule.RunningTask", "running"),
        ]

        jobs_dir = workspace_path / "jobs"
        for job_id, task_id, state in jobs_data:
            JobModel.create(
                job_id=job_id,
                experiment_id="test_exp",
                run_id="run_001",
                task_id=task_id,
                locator=job_id,
                state=state,
                submitted_time=time.time(),
            )
            # Create job directories
            job_dir = jobs_dir / task_id / job_id
            job_dir.mkdir(parents=True, exist_ok=True)
            # Create marker files based on state
            script_name = task_id.rsplit(".", 1)[-1]
            if state == "done":
                (job_dir / f"{script_name}.done").touch()
            elif state == "error":
                (job_dir / f"{script_name}.failed").write_text("{}")

    close_workspace_database(db)

    yield workspace_path


def test_jobs_list_empty_workspace(workspace_path):
    """Test jobs list on empty workspace"""
    runner = CliRunner()
    result = runner.invoke(cli, ["jobs", "--workdir", str(workspace_path), "list"])

    assert result.exit_code == 0
    assert "No jobs found" in result.output


def test_jobs_list_with_jobs(workspace_with_jobs):
    """Test jobs list shows all jobs"""
    runner = CliRunner()
    result = runner.invoke(cli, ["jobs", "--workdir", str(workspace_with_jobs), "list"])

    assert result.exit_code == 0
    assert "job_done_1" in result.output
    assert "job_done_2" in result.output
    assert "job_error_1" in result.output
    assert "job_running_1" in result.output
    assert "DONE" in result.output
    assert "FAIL" in result.output


def test_jobs_list_with_filter(workspace_with_jobs):
    """Test jobs list with filter expression"""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "jobs",
            "--workdir",
            str(workspace_with_jobs),
            "list",
            "--filter",
            '@state = "done"',
        ],
    )

    assert result.exit_code == 0
    assert "job_done_1" in result.output
    assert "job_done_2" in result.output
    assert "job_error_1" not in result.output
    assert "job_running_1" not in result.output


def test_jobs_list_fullpath(workspace_with_jobs):
    """Test jobs list with fullpath option"""
    runner = CliRunner()
    result = runner.invoke(
        cli, ["jobs", "--workdir", str(workspace_with_jobs), "list", "--fullpath"]
    )

    assert result.exit_code == 0
    # Should show full paths instead of task_id/job_id format
    assert str(workspace_with_jobs / "jobs") in result.output


@pytest.fixture
def workspace_with_timed_jobs(workspace_path):
    """Create a workspace with jobs that have different submission times"""
    db_path = workspace_path / ".experimaestro" / "workspace.db"
    db = initialize_workspace_database(db_path, read_only=False)

    with db.bind_ctx(ALL_MODELS):
        ExperimentModel.create(experiment_id="test_exp", current_run_id="run_001")
        ExperimentRunModel.create(
            experiment_id="test_exp", run_id="run_001", status="active"
        )

        # Create jobs with different submission times (oldest to newest)
        base_time = time.time()
        jobs_data = [
            ("job_oldest", "mymodule.Task", "done", base_time - 3600),  # 1 hour ago
            ("job_middle", "mymodule.Task", "done", base_time - 1800),  # 30 min ago
            ("job_newest", "mymodule.Task", "done", base_time),  # now
        ]

        jobs_dir = workspace_path / "jobs"
        for job_id, task_id, state, submit_time in jobs_data:
            JobModel.create(
                job_id=job_id,
                experiment_id="test_exp",
                run_id="run_001",
                task_id=task_id,
                locator=job_id,
                state=state,
                submitted_time=submit_time,
            )
            job_dir = jobs_dir / task_id / job_id
            job_dir.mkdir(parents=True, exist_ok=True)
            script_name = task_id.rsplit(".", 1)[-1]
            (job_dir / f"{script_name}.done").touch()

    close_workspace_database(db)
    yield workspace_path


def test_jobs_list_sorted_by_date(workspace_with_timed_jobs):
    """Test that jobs are sorted by submission date (most recent first)"""
    runner = CliRunner()
    result = runner.invoke(
        cli, ["jobs", "--workdir", str(workspace_with_timed_jobs), "list"]
    )

    assert result.exit_code == 0
    output = result.output

    # Verify order: newest should appear before middle, middle before oldest
    newest_pos = output.find("job_newest")
    middle_pos = output.find("job_middle")
    oldest_pos = output.find("job_oldest")

    assert newest_pos < middle_pos < oldest_pos, (
        f"Jobs should be sorted by date (newest first). "
        f"Positions: newest={newest_pos}, middle={middle_pos}, oldest={oldest_pos}"
    )


def test_jobs_list_with_count(workspace_with_timed_jobs):
    """Test jobs list with --count option"""
    runner = CliRunner()
    result = runner.invoke(
        cli, ["jobs", "--workdir", str(workspace_with_timed_jobs), "list", "-c", "2"]
    )

    assert result.exit_code == 0
    output = result.output

    # Should only show 2 most recent jobs
    assert "job_newest" in output
    assert "job_middle" in output
    assert "job_oldest" not in output


def test_jobs_list_count_zero_shows_all(workspace_with_timed_jobs):
    """Test that --count 0 shows all jobs (default behavior)"""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["jobs", "--workdir", str(workspace_with_timed_jobs), "list", "--count", "0"],
    )

    assert result.exit_code == 0
    output = result.output

    # All jobs should be present
    assert "job_newest" in output
    assert "job_middle" in output
    assert "job_oldest" in output


def test_jobs_list_with_experiment_filter(workspace_with_jobs):
    """Test jobs list filtered by experiment"""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "jobs",
            "--workdir",
            str(workspace_with_jobs),
            "list",
            "--experiment",
            "test_exp",
        ],
    )

    assert result.exit_code == 0
    assert "job_done_1" in result.output

    # Test with non-existent experiment
    result = runner.invoke(
        cli,
        [
            "jobs",
            "--workdir",
            str(workspace_with_jobs),
            "list",
            "--experiment",
            "nonexistent",
        ],
    )
    assert result.exit_code == 0
    assert "No jobs found" in result.output


def test_jobs_clean_dry_run(workspace_with_jobs):
    """Test jobs clean without --perform (dry run)"""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "jobs",
            "--workdir",
            str(workspace_with_jobs),
            "clean",
            "--filter",
            '@state = "done"',
        ],
    )

    assert result.exit_code == 0
    assert "dry run" in result.output.lower()

    # Verify directories still exist
    jobs_dir = workspace_with_jobs / "jobs"
    assert (jobs_dir / "mymodule.DoneTask" / "job_done_1").exists()
    assert (jobs_dir / "mymodule.DoneTask" / "job_done_2").exists()


def test_jobs_clean_with_perform(workspace_with_jobs):
    """Test jobs clean with --perform actually deletes jobs"""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "jobs",
            "--workdir",
            str(workspace_with_jobs),
            "clean",
            "--filter",
            '@state = "done"',
            "--perform",
        ],
    )

    assert result.exit_code == 0
    assert "Cleaned" in result.output

    # Verify directories are deleted
    jobs_dir = workspace_with_jobs / "jobs"
    assert not (jobs_dir / "mymodule.DoneTask" / "job_done_1").exists()
    assert not (jobs_dir / "mymodule.DoneTask" / "job_done_2").exists()

    # Verify database entries are deleted
    provider = WorkspaceStateProvider.get_instance(workspace_with_jobs, read_only=True)
    try:
        jobs = provider.get_all_jobs()
        job_ids = [j.identifier for j in jobs]
        assert "job_done_1" not in job_ids
        assert "job_done_2" not in job_ids
        # Error and running jobs should still exist
        assert "job_error_1" in job_ids
        assert "job_running_1" in job_ids
    finally:
        provider.close()


def test_jobs_clean_does_not_clean_running(workspace_with_jobs):
    """Test that jobs clean does not clean running jobs"""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["jobs", "--workdir", str(workspace_with_jobs), "clean", "--perform"],
    )

    assert result.exit_code == 0

    # Verify running job is NOT cleaned
    provider = WorkspaceStateProvider.get_instance(workspace_with_jobs, read_only=True)
    try:
        jobs = provider.get_all_jobs()
        job_ids = [j.identifier for j in jobs]
        assert "job_running_1" in job_ids
    finally:
        provider.close()


def test_jobs_kill_dry_run(workspace_with_jobs):
    """Test jobs kill without --perform (dry run)"""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "jobs",
            "--workdir",
            str(workspace_with_jobs),
            "kill",
            "--filter",
            '@state = "running"',
        ],
    )

    assert result.exit_code == 0
    assert "dry run" in result.output.lower()


def test_jobs_kill_with_perform(workspace_with_jobs):
    """Test jobs kill with --perform actually kills jobs and updates DB"""
    import json
    from unittest.mock import MagicMock, patch

    # Create a PID file for the running job so kill can find it
    jobs_dir = workspace_with_jobs / "jobs"
    running_job_dir = jobs_dir / "mymodule.RunningTask" / "job_running_1"
    pid_file = running_job_dir / "RunningTask.pid"
    pid_file.write_text(json.dumps({"type": "local", "pid": 12345}))

    # Create a mock process
    mock_process = MagicMock()
    mock_process.kill = MagicMock()

    with patch(
        "experimaestro.connectors.Process.fromDefinition", return_value=mock_process
    ):
        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "jobs",
                "--workdir",
                str(workspace_with_jobs),
                "kill",
                "--filter",
                '@state = "running"',
                "--perform",
            ],
        )

        assert result.exit_code == 0
        assert "KILLED" in result.output

        # Verify kill was called on the mock process
        mock_process.kill.assert_called_once()

    # Verify database state was updated to error
    provider = WorkspaceStateProvider.get_instance(workspace_with_jobs, read_only=True)
    try:
        jobs = provider.get_all_jobs()
        running_job = next((j for j in jobs if j.identifier == "job_running_1"), None)
        assert running_job is not None
        # State should be updated to error after kill
        from experimaestro.scheduler import JobState

        assert running_job.state == JobState.ERROR
    finally:
        provider.close()


def test_jobs_path(workspace_with_jobs):
    """Test jobs path command"""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "jobs",
            "--workdir",
            str(workspace_with_jobs),
            "path",
            "mymodule.DoneTask/job_done_1",
        ],
    )

    assert result.exit_code == 0
    expected_path = workspace_with_jobs / "jobs" / "mymodule.DoneTask" / "job_done_1"
    assert str(expected_path) in result.output


def test_jobs_path_nonexistent(workspace_with_jobs):
    """Test jobs path command for non-existent job"""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "jobs",
            "--workdir",
            str(workspace_with_jobs),
            "path",
            "mymodule.Task/nonexistent",
        ],
    )

    assert result.exit_code == 0
    assert "not found" in result.output.lower()


def test_jobs_log_nonexistent(workspace_with_jobs):
    """Test jobs log command for non-existent log"""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "jobs",
            "--workdir",
            str(workspace_with_jobs),
            "log",
            "mymodule.DoneTask/job_done_1",
        ],
    )

    assert result.exit_code == 0
    assert "not found" in result.output.lower()


def test_sync_detects_dead_process_and_updates_disk(tmp_path):
    """Test that sync detects a dead 'running' job and updates disk state

    This tests the scenario where a job crashed without updating its state:
    - Job has a .pid file (marked as running)
    - Process is no longer running (dead)
    - Sync should detect this and create a .failed file
    """
    import json
    from unittest.mock import MagicMock, patch

    from experimaestro.scheduler.state_sync import check_process_alive

    # Create a job directory with a PID file for a non-existent process
    job_path = tmp_path / "jobs" / "mymodule.CrashedTask" / "job_crashed_1"
    job_path.mkdir(parents=True)

    scriptname = "CrashedTask"
    pid_file = job_path / f"{scriptname}.pid"
    pid_file.write_text(
        json.dumps({"type": "local", "pid": 999999})
    )  # Non-existent PID

    # Create a mock process that reports as finished/dead
    mock_process = MagicMock()
    mock_state = MagicMock()
    mock_state.finished = True
    mock_state.exitcode = 137  # Killed by signal

    # Mock aio_state to return the finished state
    async def mock_aio_state(timeout):
        return mock_state

    mock_process.aio_state = mock_aio_state

    with patch(
        "experimaestro.connectors.Process.fromDefinition", return_value=mock_process
    ):
        # Check that process is detected as dead
        is_alive = check_process_alive(job_path, scriptname, update_disk=True)

        assert is_alive is False

        # Verify .failed file was created
        failed_file = job_path / f"{scriptname}.failed"
        assert failed_file.exists(), ".failed file should be created for dead process"

        # Verify .failed file contains correct data
        failed_data = json.loads(failed_file.read_text())
        assert failed_data["exit_code"] == 137
        assert failed_data["failure_status"] == "UNKNOWN"

        # Verify .pid file was removed
        assert not pid_file.exists(), ".pid file should be removed for dead process"


def test_sync_job_state_detects_dead_process(tmp_path):
    """Test that scan_job_state_from_disk detects dead 'running' jobs"""
    import json
    from unittest.mock import MagicMock, patch

    from experimaestro.scheduler.state_sync import scan_job_state_from_disk

    # Create a job directory with only a PID file (no .done or .failed)
    job_path = tmp_path / "jobs" / "mymodule.CrashedTask" / "job_crashed_2"
    job_path.mkdir(parents=True)

    scriptname = "CrashedTask"
    pid_file = job_path / f"{scriptname}.pid"
    pid_file.write_text(json.dumps({"type": "local", "pid": 999998}))

    # Mock process as dead
    mock_process = MagicMock()
    mock_state = MagicMock()
    mock_state.finished = True
    mock_state.exitcode = 1

    async def mock_aio_state(timeout):
        return mock_state

    mock_process.aio_state = mock_aio_state

    with patch(
        "experimaestro.connectors.Process.fromDefinition", return_value=mock_process
    ):
        # Scan job state - should detect dead process
        job_state = scan_job_state_from_disk(job_path, scriptname, check_running=True)

        assert job_state is not None
        assert job_state["state"] == "error"
        assert job_state["failure_reason"] == "UNKNOWN"


def test_check_process_alive_with_lock_held(tmp_path):
    """Test that check_process_alive returns True when lock is held (job running)"""
    import json
    from unittest.mock import patch, MagicMock

    from experimaestro.scheduler.state_sync import check_process_alive

    # Create a job directory with a PID file
    job_path = tmp_path / "jobs" / "mymodule.RunningTask" / "job_locked"
    job_path.mkdir(parents=True)

    scriptname = "RunningTask"
    pid_file = job_path / f"{scriptname}.pid"
    pid_file.write_text(json.dumps({"type": "local", "pid": 12345}))

    # Mock the lock to simulate it being held by another process
    mock_lock = MagicMock()
    mock_lock.acquire.return_value = False  # Simulate lock already held

    with patch(
        "experimaestro.scheduler.state_sync.fasteners.InterProcessLock",
        return_value=mock_lock,
    ):
        # check_process_alive should return True because it can't acquire the lock
        is_alive = check_process_alive(job_path, scriptname, update_disk=True)

        assert is_alive is True, "Should assume job is running when lock is held"

        # .pid file should still exist (not removed)
        assert pid_file.exists(), ".pid file should not be removed when lock is held"

        # No .failed file should be created
        failed_file = job_path / f"{scriptname}.failed"
        assert (
            not failed_file.exists()
        ), ".failed file should not be created when lock is held"



class HandledException(Exception):
    pass


class GracefulTimeout(Exception):
    """Exception raised by ResumableTask to stop gracefully before a hard timeout.

    ResumableTasks can catch signals or monitor time remaining and raise this
    exception to stop cleanly before the scheduler (e.g., SLURM) terminates
    the process. The task will be retried if retry_count < max_retries.

    Example:
        class LongTraining(ResumableTask):
            def execute(self):
                for epoch in range(1000):
                    if time_remaining() < MIN_TIME_FOR_EPOCH:
                        raise GracefulTimeout("Not enough time for another epoch")
                    train_one_epoch()
                    save_checkpoint()
    """

    def __init__(self, message: str = "Task stopped gracefully before timeout"):
        self.message = message
        super().__init__(message)



from .config_walk import ConfigWalkContext, ConfigWalk
from .config import (
    ConfigMixin,
    Config,
    InstanceConfig,
    ConfigInformation,
    Task,
    ResumableTask,
    LightweightTask,
    WatchedOutput,
    DependentMarker,
    copyconfig,
    setmeta,
    cache,
    logger,
)

from .config_utils import (
    getqualattr,
    add_to_path,
    ObjectStore,
    SealedError,
    TaggedValue,
)


__all__ = [
    "ConfigMixin",
    "Config",
    "InstanceConfig",
    "ConfigInformation",
    "ConfigWalkContext",
    "ConfigWalk",
    "Task",
    "ResumableTask",
    "LightweightTask",
    "ObjectStore",
    "WatchedOutput",
    "SealedError",
    "DependentMarker",
    "TaggedValue",
    "getqualattr",
    "copyconfig",
    "setmeta",
    "cache",
    "add_to_path",
    "logger",
]



"""Main Textual TUI application for experiment monitoring"""

import logging
from pathlib import Path
from typing import Optional
from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.widgets import (
    Header,
    Footer,
    DataTable,
    Label,
    TabbedContent,
    TabPane,
    RichLog,
    Button,
    Static,
)
from textual.widget import Widget
from textual.reactive import reactive
from textual.binding import Binding
from textual.message import Message
from textual.screen import ModalScreen
from textual import events
from experimaestro.scheduler.state_provider import (
    WorkspaceStateProvider,
    StateEvent,
    StateEventType,
)
from experimaestro.tui.log_viewer import LogViewerScreen


class QuitConfirmScreen(ModalScreen[bool]):
    """Modal screen for quit confirmation"""

    CSS = """
    QuitConfirmScreen {
        align: center middle;
    }

    #quit-dialog {
        width: 60;
        height: auto;
        border: thick $primary;
        background: $surface;
        padding: 1 2;
    }

    #quit-title {
        text-align: center;
        text-style: bold;
        margin-bottom: 1;
    }

    #quit-message {
        margin-bottom: 1;
    }

    #quit-warning {
        color: $warning;
        text-style: bold;
        margin-bottom: 1;
    }

    #quit-buttons {
        width: 100%;
        height: auto;
        align: center middle;
    }

    #quit-buttons Button {
        margin: 0 1;
    }
    """

    def __init__(self, has_active_experiment: bool = False):
        super().__init__()
        self.has_active_experiment = has_active_experiment

    def compose(self) -> ComposeResult:
        with Vertical(id="quit-dialog"):
            yield Static("Quit Experimaestro?", id="quit-title")

            if self.has_active_experiment:
                yield Static(
                    "⚠️  The experiment is still in progress.\n"
                    "Quitting will prevent new jobs from being launched.",
                    id="quit-warning",
                )
            else:
                yield Static("Are you sure you want to quit?", id="quit-message")

            with Horizontal(id="quit-buttons"):
                yield Button("Quit", variant="error", id="quit-yes")
                yield Button("Cancel", variant="primary", id="quit-no")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "quit-yes":
            self.dismiss(True)
        else:
            self.dismiss(False)


def get_status_text(status: str):
    if status == "done":
        status_text = "✓ Done"
    elif status == "error":
        status_text = "❌ Failed"
    elif status == "running":
        status_text = "⏳ Running"
    elif status == "unscheduled":
        status_text = "📆 Unscheduled"
    else:
        status_text = status

    return status_text


class CaptureLog(RichLog):
    """Custom RichLog that handles Print events and logging"""

    def on_mount(self) -> None:
        """Enable print capturing when widget is mounted"""
        # Capture print statements
        self.begin_capture_print()

    def on_print(self, event: events.Print) -> None:
        """Handle print events from captured stdout/stderr"""
        self.write(event.text)


class ExperimentsList(Widget):
    """Widget displaying list of experiments"""

    current_experiment: reactive[Optional[str]] = reactive(None)
    collapsed: reactive[bool] = reactive(False)

    def __init__(self, state_provider: WorkspaceStateProvider) -> None:
        super().__init__()
        self.state_provider = state_provider
        self.experiments = []

    def compose(self) -> ComposeResult:
        # Collapsed header (hidden initially)
        with Horizontal(id="collapsed-header", classes="hidden"):
            yield Label("", id="collapsed-experiment-info")

        # Full experiments table
        with Container(id="experiments-table-container"):
            yield Label("Experiments", classes="section-title")
            yield DataTable(id="experiments-table", cursor_type="row")

    def on_mount(self) -> None:
        """Initialize the experiments table"""
        table = self.query_one("#experiments-table", DataTable)
        table.add_column("ID", key="id")
        table.add_column("Jobs", key="jobs")
        table.add_column("Status", key="status")
        self.refresh_experiments()

        # If there's only one experiment, automatically select it
        if len(self.experiments) == 1:
            exp_id = self.experiments[0].experiment_id
            self.current_experiment = exp_id
            self.collapse_to_experiment(exp_id)
            self.post_message(ExperimentSelected(exp_id))

    def refresh_experiments(self) -> None:
        """Refresh the experiments list from state provider"""
        table = self.query_one("#experiments-table", DataTable)

        try:
            self.experiments = self.state_provider.get_experiments()
            self.log(
                f"Refreshing experiments: found {len(self.experiments)} experiments"
            )
        except Exception as e:
            self.log(f"ERROR refreshing experiments: {e}")
            import traceback

            self.log(traceback.format_exc())
            self.experiments = []
            return

        # Get existing row keys
        existing_keys = set(table.rows.keys())
        current_exp_ids = set()

        for exp in self.experiments:
            exp_id = exp.experiment_id
            current_exp_ids.add(exp_id)
            total = exp.total_jobs
            finished = exp.finished_jobs
            failed = exp.failed_jobs

            # Determine status
            if failed > 0:
                status = f"❌ {failed} failed"
            elif finished == total and total > 0:
                status = "✓ Done"
            elif finished < total:
                status = f"⏳ {finished}/{total}"
            else:
                status = "Empty"

            jobs_text = f"{finished}/{total}"

            # Update existing row or add new one
            if exp_id in existing_keys:
                table.update_cell(exp_id, "id", exp_id, update_width=True)
                table.update_cell(exp_id, "jobs", jobs_text, update_width=True)
                table.update_cell(exp_id, "status", status, update_width=True)
            else:
                table.add_row(exp_id, jobs_text, status, key=exp_id)

        # Remove rows for experiments that no longer exist
        for old_exp_id in existing_keys - current_exp_ids:
            table.remove_row(old_exp_id)

    def on_data_table_row_selected(self, event: DataTable.RowSelected) -> None:
        """Handle experiment selection"""
        if event.row_key:
            self.current_experiment = str(event.row_key.value)
            self.collapse_to_experiment(self.current_experiment)
            self.post_message(ExperimentSelected(str(event.row_key.value)))

    def collapse_to_experiment(self, experiment_id: str) -> None:
        """Collapse the experiments list to show only the selected experiment"""
        # Find experiment info
        exp_info = next(
            (exp for exp in self.experiments if exp.experiment_id == experiment_id),
            None,
        )
        if not exp_info:
            return

        # Update collapsed header
        total = exp_info.total_jobs
        finished = exp_info.finished_jobs
        failed = exp_info.failed_jobs

        if failed > 0:
            status = f"❌ {failed} failed"
        elif finished == total and total > 0:
            status = "✓ Done"
        elif finished < total:
            status = f"⏳ {finished}/{total}"
        else:
            status = "Empty"

        collapsed_label = self.query_one("#collapsed-experiment-info", Label)
        collapsed_label.update(f"📊 {experiment_id} - {status} (click to expand)")

        # Hide table, show collapsed header
        self.query_one("#experiments-table-container").add_class("hidden")
        self.query_one("#collapsed-header").remove_class("hidden")
        self.collapsed = True

    def expand_experiments(self) -> None:
        """Expand back to full experiments list"""
        # Show table, hide collapsed header
        self.query_one("#collapsed-header").add_class("hidden")
        self.query_one("#experiments-table-container").remove_class("hidden")
        self.collapsed = False
        self.current_experiment = None

        # Focus the experiments table
        table = self.query_one("#experiments-table", DataTable)
        table.focus()

    def on_click(self) -> None:
        """Handle clicks on the widget"""
        if self.collapsed:
            self.expand_experiments()
            self.post_message(ExperimentDeselected())


class ExperimentSelected(Message):
    """Message sent when an experiment is selected"""

    def __init__(self, experiment_id: str) -> None:
        super().__init__()
        self.experiment_id = experiment_id


class ExperimentDeselected(Message):
    """Message sent when an experiment is deselected"""

    pass


class JobSelected(Message):
    """Message sent when a job is selected"""

    def __init__(self, job_id: str, experiment_id: str) -> None:
        super().__init__()
        self.job_id = job_id
        self.experiment_id = experiment_id


class JobDeselected(Message):
    """Message sent when returning from job detail view"""

    pass


class ViewJobLogs(Message):
    """Message sent when user wants to view job logs"""

    def __init__(self, job_path: str, task_id: str) -> None:
        super().__init__()
        self.job_path = job_path
        self.task_id = task_id


class JobDetailView(Widget):
    """Widget displaying detailed job information"""

    BINDINGS = [
        Binding("l", "view_logs", "View Logs", priority=True),
    ]

    def __init__(self, state_provider: WorkspaceStateProvider) -> None:
        super().__init__()
        self.state_provider = state_provider
        self.current_job_id: Optional[str] = None
        self.current_experiment_id: Optional[str] = None
        self.job_data: Optional[dict] = None

    def compose(self) -> ComposeResult:
        yield Label("Job Details", classes="section-title")
        with Vertical(id="job-detail-content"):
            yield Label("", id="job-id-label")
            yield Label("", id="job-task-label")
            yield Label("", id="job-status-label")
            yield Label("", id="job-path-label")
            yield Label("", id="job-times-label")
            yield Label("Tags:", classes="subsection-title")
            yield Label("", id="job-tags-label")
            yield Label("Progress:", classes="subsection-title")
            yield Label("", id="job-progress-label")
            yield Label("", id="job-logs-hint")

    def action_view_logs(self) -> None:
        """View job logs with toolong"""
        if self.job_data and self.job_data.path and self.job_data.task_id:
            self.post_message(
                ViewJobLogs(str(self.job_data.path), self.job_data.task_id)
            )

    def set_job(self, job_id: str, experiment_id: str) -> None:
        """Set the job to display"""
        self.current_job_id = job_id
        self.current_experiment_id = experiment_id
        self.refresh_job_detail()

    def refresh_job_detail(self) -> None:
        """Refresh job details from state provider"""
        if not self.current_job_id or not self.current_experiment_id:
            return

        job = self.state_provider.get_job(
            self.current_job_id, self.current_experiment_id
        )
        if not job:
            self.log(f"Job not found: {self.current_job_id}")
            return

        self.job_data = job

        # Update labels
        self.query_one("#job-id-label", Label).update(f"Job ID: {job.identifier}")
        self.query_one("#job-task-label", Label).update(f"Task: {job.task_id}")

        # Format status
        status_text = get_status_text(job.state.name if job.state else "unknown")

        self.query_one("#job-status-label", Label).update(f"Status: {status_text}")

        # Path (from locator)
        locator = job.locator or "-"
        self.query_one("#job-path-label", Label).update(f"Locator: {locator}")

        # Times - format timestamps
        from datetime import datetime
        import time as time_module

        def format_time(ts):
            if ts:
                return datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H:%M:%S")
            return "-"

        def format_duration(seconds: float) -> str:
            if seconds < 0:
                return "-"
            seconds = int(seconds)
            if seconds < 60:
                return f"{seconds}s"
            elif seconds < 3600:
                return f"{seconds // 60}m {seconds % 60}s"
            elif seconds < 86400:
                return f"{seconds // 3600}h {(seconds % 3600) // 60}m"
            else:
                return f"{seconds // 86400}d {(seconds % 86400) // 3600}h"

        submitted = format_time(job.submittime)
        start = format_time(job.starttime)
        end = format_time(job.endtime)

        # Calculate duration
        duration = "-"
        if job.starttime:
            if job.endtime:
                duration = format_duration(job.endtime - job.starttime)
            else:
                duration = (
                    format_duration(time_module.time() - job.starttime) + " (running)"
                )

        times_text = f"Submitted: {submitted} | Start: {start} | End: {end} | Duration: {duration}"
        self.query_one("#job-times-label", Label).update(times_text)

        # Tags - job.tags is now a dict
        tags = job.tags
        if tags:
            tags_text = ", ".join(f"{k}={v}" for k, v in tags.items())
        else:
            tags_text = "(no tags)"
        self.query_one("#job-tags-label", Label).update(tags_text)

        # Progress
        progress_list = job.progress or []
        if progress_list:
            progress_lines = []
            for p in progress_list:
                level = p.get("level", 0)
                pct = p.get("progress", 0) * 100
                desc = p.get("desc", "")
                indent = "  " * level
                progress_lines.append(f"{indent}{pct:.1f}% {desc}")
            progress_text = "\n".join(progress_lines) if progress_lines else "-"
        else:
            progress_text = "-"
        self.query_one("#job-progress-label", Label).update(progress_text)

        # Log files hint - log files are named after the last part of the task ID
        job_path = job.path
        task_id = job.task_id
        if job_path and task_id:
            # Extract the last component of the task ID (e.g., "evaluate" from "mnist_xp.learn.evaluate")
            task_name = task_id.split(".")[-1]
            stdout_path = job_path / f"{task_name}.out"
            stderr_path = job_path / f"{task_name}.err"
            logs_exist = stdout_path.exists() or stderr_path.exists()
            if logs_exist:
                self.query_one("#job-logs-hint", Label).update(
                    "[bold cyan]Press 'l' to view logs[/bold cyan]"
                )
            else:
                self.query_one("#job-logs-hint", Label).update("(no log files found)")
        else:
            self.query_one("#job-logs-hint", Label).update("")


class JobsTable(Widget):
    """Widget displaying jobs for selected experiment"""

    def __init__(self, state_provider: WorkspaceStateProvider) -> None:
        super().__init__()
        self.state_provider = state_provider
        self.current_experiment: Optional[str] = None

    def compose(self) -> ComposeResult:
        yield Label("Jobs", classes="section-title")
        yield DataTable(id="jobs-table", cursor_type="row")

    def on_mount(self) -> None:
        """Initialize the jobs table"""
        table = self.query_one("#jobs-table", DataTable)
        table.add_column("Job ID", key="job_id", width=8)
        table.add_column("Task", key="task")
        table.add_column("Status", key="status")
        table.add_column("Progress", key="progress")
        table.add_column("Submitted", key="submitted")
        table.add_column("Duration", key="duration")
        table.cursor_type = "row"

    def set_experiment(self, experiment_id: Optional[str]) -> None:
        """Set the current experiment and refresh jobs"""
        self.current_experiment = experiment_id
        self.refresh_jobs()

    def refresh_jobs(self) -> None:
        """Refresh the jobs list from state provider"""
        table = self.query_one("#jobs-table", DataTable)

        if not self.current_experiment:
            return

        jobs = self.state_provider.get_jobs(self.current_experiment)
        self.log(f"Refreshing jobs for {self.current_experiment}: {len(jobs)} jobs")

        # Sort jobs by submission time (oldest first)
        # Jobs without submittime go to the end
        jobs.sort(key=lambda j: j.submittime or float("inf"))

        # Get existing row keys
        existing_keys = set(table.rows.keys())
        current_job_ids = set()

        for job in jobs:
            job_id = job.identifier
            current_job_ids.add(job_id)
            task_id = job.task_id
            status = job.state.name if job.state else "unknown"

            # Format status with icon
            status_text = get_status_text(status)

            # Format progress - only show for running jobs
            if job.state and job.state.finished():
                # Don't show progress for finished jobs
                progress_text = "-"
            else:
                progress_list = job.progress or []
                if progress_list:
                    # Get the last progress entry
                    last_progress = progress_list[-1]
                    progress_pct = last_progress.get("progress", 0) * 100
                    progress_text = f"{progress_pct:.0f}%"
                else:
                    progress_text = "-"

            # Format timestamps
            from datetime import datetime
            import time as time_module

            submitted = "-"
            if job.submittime:
                submitted = datetime.fromtimestamp(job.submittime).strftime(
                    "%Y-%m-%d %H:%M"
                )

            # Calculate duration
            start = job.starttime
            end = job.endtime
            duration = "-"
            if start:
                if end:
                    # Job finished - show total duration
                    elapsed = end - start
                else:
                    # Job still running - show elapsed time so far
                    elapsed = time_module.time() - start
                # Format duration as human-readable
                duration = self._format_duration(elapsed)

            row_data = (
                job_id,
                task_id,
                status_text,
                progress_text,
                submitted,
                duration,
            )

            # Update existing row or add new one
            if job_id in existing_keys:
                table.update_cell(job_id, "job_id", job_id, update_width=True)
                table.update_cell(job_id, "task", task_id, update_width=True)
                table.update_cell(job_id, "status", status_text, update_width=True)
                table.update_cell(job_id, "progress", progress_text, update_width=True)
                table.update_cell(job_id, "submitted", submitted, update_width=True)
                table.update_cell(job_id, "duration", duration, update_width=True)
                self.log(f"Updated job row: {job_id}")
            else:
                table.add_row(*row_data, key=job_id)
                self.log(f"Added job row: {job_id}")

        # Remove rows for jobs that no longer exist
        for old_job_id in existing_keys - current_job_ids:
            table.remove_row(old_job_id)

        self.log(f"Jobs table now has {table.row_count} rows")

    def _format_duration(self, seconds: float) -> str:
        """Format duration in seconds to human-readable string"""
        if seconds < 0:
            return "-"

        seconds = int(seconds)
        if seconds < 60:
            return f"{seconds}s"
        elif seconds < 3600:
            minutes = seconds // 60
            secs = seconds % 60
            return f"{minutes}m {secs}s"
        elif seconds < 86400:
            hours = seconds // 3600
            minutes = (seconds % 3600) // 60
            return f"{hours}h {minutes}m"
        else:
            days = seconds // 86400
            hours = (seconds % 86400) // 3600
            return f"{days}d {hours}h"

    def on_data_table_row_selected(self, event: DataTable.RowSelected) -> None:
        """Handle job selection"""
        if event.row_key and self.current_experiment:
            job_id = str(event.row_key.value)
            self.post_message(JobSelected(job_id, self.current_experiment))


class ExperimentTUI(App):
    """Textual TUI for monitoring experiments"""

    CSS = """
    #main-container {
        width: 100%;
        height: 100%;
    }

    ExperimentsList {
        width: 100%;
        height: auto;
    }

    Monitor {
        height: 100%;
    }

    #experiments-table-container {
        width: 100%;
        height: auto;
    }

    #collapsed-header {
        background: $boost;
        padding: 1;
        text-style: bold;
        border: solid green;
        height: auto;
        width: 100%;
    }

    #collapsed-header:hover {
        background: $primary;
    }

    #collapsed-experiment-info {
        width: 100%;
        height: auto;
    }

    #jobs-container {
        width: 100%;
        height: 1fr;
        border: solid blue;
        padding: 1;
    }

    JobsTable {
        width: 100%;
        height: 100%;
    }

    #jobs-table {
        height: 100%;
    }

    .hidden {
        display: none;
    }

    .section-title {
        background: $boost;
        padding: 1;
        text-style: bold;
    }

    #experiments-table {
        height: auto;
        min-height: 5;
    }

    DataTable {
        height: 100%;
    }

    RichLog {
        height: 100%;
        border: solid cyan;
    }

    TabbedContent {
        height: 100%;
    }

    #job-detail-container {
        width: 100%;
        height: 1fr;
        border: solid magenta;
        padding: 1;
    }

    JobDetailView {
        width: 100%;
        height: 100%;
    }

    #job-detail-content {
        padding: 1;
    }

    #job-detail-content Label {
        margin-bottom: 1;
    }

    .subsection-title {
        background: $surface;
        padding: 0 1;
        text-style: italic;
        margin-top: 1;
    }

    #job-logs-hint {
        margin-top: 1;
    }
    """

    BINDINGS = [
        Binding("q", "quit", "Quit", priority=True),
        Binding("r", "refresh", "Refresh", priority=True),
        Binding("escape", "go_back", "Back", priority=True),
        Binding("l", "view_logs", "View Logs"),
    ]

    def __init__(
        self,
        workdir: Optional[Path] = None,
        watch: bool = True,
        state_provider: Optional[WorkspaceStateProvider] = None,
        show_logs: bool = False,
    ):
        """Initialize the TUI

        Args:
            workdir: Workspace directory (required if state_provider not provided)
            watch: Enable filesystem watching for workspace mode
            state_provider: Pre-initialized state provider (for active experiments)
            show_logs: Whether to show the logs tab (for active experiments)
        """
        super().__init__()
        self.workdir = workdir
        self.watch = watch
        self.show_logs = show_logs
        self._listener_registered = False

        # Initialize state provider before compose
        if state_provider:
            self.state_provider = state_provider
            self.owns_provider = False  # Don't close external provider
            self._has_active_experiment = True  # External provider = active experiment
        else:
            from experimaestro.scheduler.state_provider import WorkspaceStateProvider

            # Get singleton provider instance for this workspace
            self.state_provider = WorkspaceStateProvider.get_instance(
                self.workdir,
                read_only=False,
                sync_on_start=True,
                sync_interval_minutes=5,
            )
            self.owns_provider = False  # Provider is singleton, don't close
            self._has_active_experiment = False  # Just viewing, no active experiment

    def compose(self) -> ComposeResult:
        """Compose the TUI layout"""
        yield Header()

        if self.show_logs:
            # Tabbed layout with logs
            with TabbedContent():
                with TabPane("Monitor", id="monitor-tab"):
                    yield from self._compose_monitor_view()
                with TabPane("Logs", id="logs-tab"):
                    yield CaptureLog(id="logs", wrap=True, highlight=True, markup=True)
        else:
            # Simple layout without logs
            with Vertical(id="main-container"):
                yield from self._compose_monitor_view()

        yield Footer()

    def _compose_monitor_view(self):
        """Compose the monitor view with experiments, jobs, and job details"""
        yield ExperimentsList(self.state_provider)
        # Jobs view (hidden initially)
        with Vertical(id="jobs-container", classes="hidden"):
            yield JobsTable(self.state_provider)
        # Job detail view (hidden initially)
        with Vertical(id="job-detail-container", classes="hidden"):
            yield JobDetailView(self.state_provider)

    def on_mount(self) -> None:
        """Initialize the application"""
        # Resets logging
        logging.basicConfig(level=logging.INFO, force=True)

        # Get the widgets
        experiments_list = self.query_one(ExperimentsList)
        experiments_list.refresh_experiments()

        # Register as listener for push notifications from state provider
        if self.state_provider:
            self.state_provider.add_listener(self._on_state_event)
            self._listener_registered = True
            self.log("Registered state listener for push notifications")

    def _on_state_event(self, event: StateEvent) -> None:
        """Handle state change events from the state provider

        This is called from the state provider's thread, so we use
        call_from_thread to safely update the UI.
        """
        self.call_from_thread(self._handle_state_event, event)

    def _handle_state_event(self, event: StateEvent) -> None:
        """Process state event on the main thread"""
        self.log(f"State event received: {event.event_type.name}")

        if event.event_type == StateEventType.EXPERIMENT_UPDATED:
            # Refresh experiments list
            experiments_list = self.query_one(ExperimentsList)
            experiments_list.refresh_experiments()

        elif event.event_type == StateEventType.JOB_UPDATED:
            # Refresh jobs table if we're viewing the affected experiment
            jobs_table = self.query_one(JobsTable)
            event_exp_id = event.data.get("experimentId")

            if jobs_table.current_experiment == event_exp_id:
                jobs_table.refresh_jobs()

            # Also refresh job detail if we're viewing the affected job
            job_detail_container = self.query_one("#job-detail-container")
            if not job_detail_container.has_class("hidden"):
                job_detail_view = self.query_one(JobDetailView)
                event_job_id = event.data.get("jobId")
                if job_detail_view.current_job_id == event_job_id:
                    job_detail_view.refresh_job_detail()

            # Also update the experiment stats in the experiments list
            experiments_list = self.query_one(ExperimentsList)
            experiments_list.refresh_experiments()

        elif event.event_type == StateEventType.RUN_UPDATED:
            # Refresh experiments list to show updated run info
            experiments_list = self.query_one(ExperimentsList)
            experiments_list.refresh_experiments()

    def on_experiment_selected(self, message: ExperimentSelected) -> None:
        """Handle experiment selection - show jobs view"""
        self.log(f"Experiment selected: {message.experiment_id}")
        jobs_table_widget = self.query_one(JobsTable)
        jobs_table_widget.set_experiment(message.experiment_id)

        # Show jobs container
        jobs_container = self.query_one("#jobs-container")
        self.log(f"Jobs container before: hidden={jobs_container.has_class('hidden')}")
        jobs_container.remove_class("hidden")
        self.log(f"Jobs container after: hidden={jobs_container.has_class('hidden')}")

        # Focus the jobs table
        jobs_table = self.query_one("#jobs-table", DataTable)
        jobs_table.focus()

    def on_experiment_deselected(self, message: ExperimentDeselected) -> None:
        """Handle experiment deselection - hide jobs view"""
        jobs_container = self.query_one("#jobs-container")
        jobs_container.add_class("hidden")
        # Also hide job detail if visible
        job_detail_container = self.query_one("#job-detail-container")
        job_detail_container.add_class("hidden")

    def on_job_selected(self, message: JobSelected) -> None:
        """Handle job selection - show job detail view"""
        self.log(f"Job selected: {message.job_id} from {message.experiment_id}")

        # Hide jobs table, show job detail
        jobs_container = self.query_one("#jobs-container")
        jobs_container.add_class("hidden")

        job_detail_container = self.query_one("#job-detail-container")
        job_detail_container.remove_class("hidden")

        # Set the job to display
        job_detail_view = self.query_one(JobDetailView)
        job_detail_view.set_job(message.job_id, message.experiment_id)

    def on_job_deselected(self, message: JobDeselected) -> None:
        """Handle job deselection - go back to jobs view"""
        # Hide job detail, show jobs table
        job_detail_container = self.query_one("#job-detail-container")
        job_detail_container.add_class("hidden")

        jobs_container = self.query_one("#jobs-container")
        jobs_container.remove_class("hidden")

        # Focus the jobs table
        jobs_table = self.query_one("#jobs-table", DataTable)
        jobs_table.focus()

    def action_refresh(self) -> None:
        """Manually refresh the data"""
        experiments_list = self.query_one(ExperimentsList)
        jobs_table = self.query_one(JobsTable)

        experiments_list.refresh_experiments()
        jobs_table.refresh_jobs()

        # Also refresh job detail if visible
        job_detail_container = self.query_one("#job-detail-container")
        if not job_detail_container.has_class("hidden"):
            job_detail_view = self.query_one(JobDetailView)
            job_detail_view.refresh_job_detail()

    def action_go_back(self) -> None:
        """Go back one level in the navigation hierarchy"""
        # Check if job detail is visible -> go back to jobs
        job_detail_container = self.query_one("#job-detail-container")
        if not job_detail_container.has_class("hidden"):
            self.post_message(JobDeselected())
            return

        # Check if jobs list is visible -> go back to experiments
        jobs_container = self.query_one("#jobs-container")
        if not jobs_container.has_class("hidden"):
            experiments_list = self.query_one(ExperimentsList)
            if experiments_list.collapsed:
                experiments_list.expand_experiments()
                jobs_container.add_class("hidden")

    def action_view_logs(self) -> None:
        """View logs for the current job (if job detail is visible)"""
        job_detail_container = self.query_one("#job-detail-container")
        if not job_detail_container.has_class("hidden"):
            job_detail_view = self.query_one(JobDetailView)
            job_detail_view.action_view_logs()

    def on_view_job_logs(self, message: ViewJobLogs) -> None:
        """Handle request to view job logs - push LogViewerScreen"""
        job_path = Path(message.job_path)
        # Log files are named after the last part of the task ID
        task_name = message.task_id.split(".")[-1]
        stdout_path = job_path / f"{task_name}.out"
        stderr_path = job_path / f"{task_name}.err"

        # Collect existing log files
        log_files = []
        if stdout_path.exists():
            log_files.append(str(stdout_path))
        if stderr_path.exists():
            log_files.append(str(stderr_path))

        if not log_files:
            self.notify(
                f"No log files found: {task_name}.out/.err in {job_path}",
                severity="warning",
            )
            return

        # Push the log viewer screen
        job_id = job_path.name
        self.push_screen(LogViewerScreen(log_files, job_id))

    def action_switch_focus(self) -> None:
        """Switch focus between experiments and jobs tables"""
        focused = self.focused
        if focused:
            experiments_table = self.query_one("#experiments-table", DataTable)
            jobs_table = self.query_one("#jobs-table", DataTable)

            if focused == experiments_table:
                jobs_table.focus()
            else:
                experiments_table.focus()

    def action_quit(self) -> None:
        """Show quit confirmation dialog"""

        def handle_quit_response(confirmed: bool) -> None:
            if confirmed:
                self.exit()

        self.push_screen(
            QuitConfirmScreen(has_active_experiment=self._has_active_experiment),
            handle_quit_response,
        )

    def on_unmount(self) -> None:
        """Clean up when closing"""
        # Unregister listener
        if self._listener_registered and self.state_provider:
            self.state_provider.remove_listener(self._on_state_event)
            self._listener_registered = False
            self.log("Unregistered state listener")

        # Only close state provider if we own it (not external/active experiment)
        if self.state_provider and self.owns_provider:
            self.state_provider.close()



"""Tests for deprecation mechanism.

Tests cover:
- Legacy @deprecate pattern (deprecated class inherits from new class)
- New @deprecate(Target) pattern with explicit target
- @deprecate(Target, replace=True) for immediate conversion
- Deprecated attributes
- fix_deprecated CLI for migrating job directories
- Task deprecation with fix_deprecated symlink creation
"""

import logging
from typing import List

from experimaestro import field, Config, Param, Task, deprecate
from experimaestro.core.identifier import IdentifierComputer
from experimaestro.scheduler.workspace import RunMode
from experimaestro.tools.jobs import fix_deprecated

from .utils import TemporaryExperiment


def assert_equal(a, b, message=""):
    """Assert two configs have the same identifier."""
    a_id = IdentifierComputer.compute(a)
    b_id = IdentifierComputer.compute(b)
    assert a_id == b_id, f"{message}: {a_id} != {b_id}"


def assert_notequal(a, b, message=""):
    """Assert two configs have different identifiers."""
    a_id = IdentifierComputer.compute(a)
    b_id = IdentifierComputer.compute(b)
    assert a_id != b_id, f"{message}: {a_id} == {b_id}"


# --- Legacy deprecation tests ---


def test_deprecated_class_legacy():
    """Test legacy @deprecate pattern where deprecated class inherits from new class."""

    class NewConfig(Config):
        __xpmid__ = "test.deprecated.legacy.new"

    @deprecate
    class OldConfig(NewConfig):
        __xpmid__ = "test.deprecated.legacy.old"

    class DerivedConfig(NewConfig):
        __xpmid__ = "test.deprecated.legacy.derived"

    assert_notequal(
        NewConfig.C(), DerivedConfig.C(), "A derived configuration has another ID"
    )
    assert_equal(
        NewConfig.C(),
        OldConfig.C(),
        "Deprecated and new configuration have the same ID",
    )


def test_deprecated_attribute():
    """Test deprecating a parameter with a conversion method."""

    class Values(Config):
        __xpmid__ = "test.deprecated.attribute"
        values: Param[List[int]] = field(ignore_default=[])

        @deprecate
        def value(self, x):
            self.values = [x]

    assert_equal(Values.C(values=[1]), Values.C(value=1))


# --- New @deprecate(Target) pattern tests ---


def test_deprecate_with_explicit_target():
    """Test @deprecate(TargetConfig) with explicit target class."""

    class NewConfig(Config):
        __xpmid__ = "test.deprecate.explicit.new"
        value: Param[int]

    @deprecate(NewConfig)
    class OldConfig(Config):
        __xpmid__ = "test.deprecate.explicit.old"
        value: Param[int]

        def __convert__(self):
            return NewConfig.C(value=self.value)

    # Creating OldConfig returns OldConfig (not NewConfig)
    old = OldConfig.C(value=42)
    assert type(old).__name__ == "OldConfig.XPMConfig"

    # But identifiers should match because __convert__ is used for ID computation
    new = NewConfig.C(value=42)
    assert_equal(old, new)


def test_deprecate_with_parameter_transformation():
    """Test __convert__ that transforms parameters."""

    class NewConfig(Config):
        __xpmid__ = "test.deprecate.transform.new"
        values: Param[List[int]]

    @deprecate(NewConfig)
    class OldConfig(Config):
        __xpmid__ = "test.deprecate.transform.old"
        value: Param[int]

        def __convert__(self):
            # Convert single value to list
            return NewConfig.C(values=[self.value])

    # Create old config - stays as OldConfig
    old = OldConfig.C(value=42)
    assert type(old).__name__ == "OldConfig.XPMConfig"
    assert old.value == 42

    # Identifier should match the equivalent NewConfig
    new = NewConfig.C(values=[42])
    assert_equal(old, new)


def test_deprecate_chained_versions():
    """Test chained deprecation: A_v0 -> A_v1 -> A

    Each deprecated version computes its identifier via __convert__ chain.
    """

    class A(Config):
        __xpmid__ = "test.deprecate.chained.a"
        data: Param[List[str]]

    @deprecate(A)
    class A_v1(Config):
        __xpmid__ = "test.deprecate.chained.a_v1"
        items: Param[List[str]]

        def __convert__(self):
            return A.C(data=self.items)

    @deprecate(A_v1)
    class A_v0(Config):
        __xpmid__ = "test.deprecate.chained.a_v0"
        item: Param[str]

        def __convert__(self):
            return A_v1.C(items=[self.item])

    # Create configs - each stays as its own type
    v0 = A_v0.C(item="hello")
    v1 = A_v1.C(items=["hello"])
    current = A.C(data=["hello"])

    assert type(v0).__name__ == "A_v0.XPMConfig"
    assert type(v1).__name__ == "A_v1.XPMConfig"
    assert type(current).__name__ == "A.XPMConfig"

    # All should have the same identifier (computed via conversion chain)
    assert_equal(v0, v1)
    assert_equal(v1, current)
    assert_equal(v0, current)


def test_deprecate_preserves_original_identifier():
    """Test that deprecated identifier is preserved for migration (fix_deprecated)."""

    class NewStyle(Config):
        __xpmid__ = "test.deprecate.preserve.new"
        x: Param[int]

    @deprecate(NewStyle)
    class OldStyle(Config):
        __xpmid__ = "test.deprecate.preserve.old"
        x: Param[int]

        def __convert__(self):
            return NewStyle.C(x=self.x)

    xpmtype = OldStyle.__getxpmtype__()

    # The deprecated identifier should be preserved for migration
    assert xpmtype._deprecated_identifier is not None
    assert xpmtype._deprecated_identifier.name == "test.deprecate.preserve.old"

    # The target identifier should be accessible
    assert xpmtype.identifier.name == "test.deprecate.preserve.new"


# --- @deprecate(Target, replace=True) tests ---


def test_deprecate_replace_basic():
    """Test @deprecate(Target, replace=True) immediately converts to new config."""

    class NewConfig(Config):
        __xpmid__ = "test.deprecate.replace.basic.new"
        values: Param[list[int]]

    @deprecate(NewConfig, replace=True)
    class OldConfig(Config):
        __xpmid__ = "test.deprecate.replace.basic.old"
        value: Param[int]

        def __convert__(self):
            return NewConfig.C(values=[self.value])

    # Creating OldConfig should return a NewConfig instance
    result = OldConfig.C(value=42)

    # The result should be the new config type
    assert type(result).__name__ == "NewConfig.XPMConfig"

    # The values should be converted
    assert result.values == [42]


def test_deprecate_replace_identifier():
    """Test that replaced configs have the same identifier as direct creation."""

    class NewConfig(Config):
        __xpmid__ = "test.deprecate.replace.id.new"
        values: Param[list[int]]

    @deprecate(NewConfig, replace=True)
    class OldConfig(Config):
        __xpmid__ = "test.deprecate.replace.id.old"
        value: Param[int]

        def __convert__(self):
            return NewConfig.C(values=[self.value])

    # Create via old and new ways
    via_old = OldConfig.C(value=42)
    via_new = NewConfig.C(values=[42])

    # Both should be NewConfig instances with same identifier
    assert type(via_old).__name__ == "NewConfig.XPMConfig"
    assert type(via_new).__name__ == "NewConfig.XPMConfig"
    assert_equal(via_old, via_new)


def test_deprecate_replace_preserves_deprecated_id():
    """Test that deprecated identifier is preserved for fix_deprecated even with replace=True."""

    class NewConfig(Config):
        __xpmid__ = "test.deprecate.replace.preserve.new"
        values: Param[list[int]]

    @deprecate(NewConfig, replace=True)
    class OldConfig(Config):
        __xpmid__ = "test.deprecate.replace.preserve.old"
        value: Param[int]

        def __convert__(self):
            return NewConfig.C(values=[self.value])

    xpmtype = OldConfig.__getxpmtype__()

    # The deprecated identifier should be preserved
    assert xpmtype._deprecated_identifier is not None
    assert xpmtype._deprecated_identifier.name == "test.deprecate.replace.preserve.old"

    # The current identifier should point to new config
    assert xpmtype.identifier.name == "test.deprecate.replace.preserve.new"


def test_deprecate_replace_deprecated_from_preserved():
    """Test that converted config preserves reference to original deprecated config."""

    class NewConfig(Config):
        __xpmid__ = "test.deprecate.replace.from.new"
        values: Param[list[int]]

    @deprecate(NewConfig, replace=True)
    class OldConfig(Config):
        __xpmid__ = "test.deprecate.replace.from.old"
        value: Param[int]

        def __convert__(self):
            return NewConfig.C(values=[self.value])

    result = OldConfig.C(value=42)

    # The converted config should have reference to original
    assert result._deprecated_from is not None
    assert type(result._deprecated_from).__name__ == "OldConfig.XPMConfig"
    assert result._deprecated_from.value == 42


# --- fix_deprecated tests ---


def test_deprecation_info_available_for_fix_deprecated():
    """Test that deprecation info is available for fix_deprecated tool.

    The fix_deprecated tool needs to know:
    1. The original (deprecated) identifier
    2. The new (target) identifier
    This test verifies this information is available via _deprecation.
    """

    class NewConfig(Config):
        __xpmid__ = "test.fixdeprecated.info.new"
        values: Param[list[int]]

    @deprecate(NewConfig)
    class OldConfig(Config):
        __xpmid__ = "test.fixdeprecated.info.old"
        value: Param[int]

        def __convert__(self):
            return NewConfig.C(values=[self.value])

    xpmtype = OldConfig.__getxpmtype__()

    # Verify deprecation info is accessible
    assert xpmtype._deprecation is not None
    assert (
        xpmtype._deprecation.original_identifier.name == "test.fixdeprecated.info.old"
    )
    assert xpmtype._deprecation.target == NewConfig
    assert xpmtype._deprecation.replace is False

    # The identifier should point to new config
    assert xpmtype.identifier.name == "test.fixdeprecated.info.new"


def test_deprecation_info_with_replace():
    """Test that deprecation info includes replace flag."""

    class NewConfig(Config):
        __xpmid__ = "test.fixdeprecated.replace.new"
        values: Param[list[int]]

    @deprecate(NewConfig, replace=True)
    class OldConfig(Config):
        __xpmid__ = "test.fixdeprecated.replace.old"
        value: Param[int]

        def __convert__(self):
            return NewConfig.C(values=[self.value])

    xpmtype = OldConfig.__getxpmtype__()

    # Verify replace flag is set
    assert xpmtype._deprecation is not None
    assert xpmtype._deprecation.replace is True
    assert (
        xpmtype._deprecation.original_identifier.name
        == "test.fixdeprecated.replace.old"
    )


# =============================================================================
# Task deprecation tests (with actual experiment context and fix_deprecated)
# =============================================================================


class NewConfigForTask(Config):
    """New configuration used in task deprecation tests."""

    __xpmid__ = "test.deprecated.task.config.new"


@deprecate
class DeprecatedConfigForTask(NewConfigForTask):
    """Deprecated configuration (legacy pattern)."""

    __xpmid__ = "test.deprecated.task.config.deprecated"


class OldConfigForTask(NewConfigForTask):
    """Old configuration without deprecate flag (for comparison)."""

    __xpmid__ = "test.deprecated.task.config.deprecated"


class TaskWithDeprecatedConfig(Task):
    """Task that uses a deprecated config as parameter."""

    __xpmid__ = "test.deprecated.task.with.config"
    p: Param[NewConfigForTask]

    def execute(self):
        pass


def _check_symlink_paths(task_new, task_old_path):
    """Helper to verify symlink was created correctly."""
    task_new_path = task_new.__xpm__.job.path  # type: Path

    assert task_new_path.exists(), f"New path {task_new_path} should exist"
    assert task_new_path.is_symlink(), f"New path {task_new_path} should be a symlink"
    assert task_new_path.resolve() == task_old_path


def test_task_deprecated_config_identifier():
    """Test that tasks using deprecated configs have correct identifiers."""
    with TemporaryExperiment("deprecated_config", maxwait=0):
        # Create tasks with new, old, and deprecated configs
        task_new = TaskWithDeprecatedConfig.C(p=NewConfigForTask.C()).submit(
            run_mode=RunMode.DRY_RUN
        )
        task_old = TaskWithDeprecatedConfig.C(p=OldConfigForTask.C()).submit(
            run_mode=RunMode.DRY_RUN
        )
        task_deprecated = TaskWithDeprecatedConfig.C(
            p=DeprecatedConfigForTask.C()
        ).submit(run_mode=RunMode.DRY_RUN)

        logging.debug("New task ID: %s", task_new.__xpm__.identifier.all.hex())
        logging.debug("Old task ID: %s", task_old.__xpm__.identifier.all.hex())
        logging.debug(
            "Deprecated task ID: %s", task_deprecated.__xpm__.identifier.all.hex()
        )

        # Old (non-deprecated) and new should have different paths
        assert (
            task_new.stdout() != task_old.stdout()
        ), "Old and new path should be different"

        # Deprecated should have same path as new (identifier matches)
        assert (
            task_new.stdout() == task_deprecated.stdout()
        ), "Deprecated path should be the same as non deprecated"


def test_task_deprecated_config_fix_deprecated():
    """Test fix_deprecated creates symlinks for tasks with deprecated configs."""
    with TemporaryExperiment("deprecated_config_fix", maxwait=0) as xp:
        task_new = TaskWithDeprecatedConfig.C(p=NewConfigForTask.C()).submit(
            run_mode=RunMode.DRY_RUN
        )

        # Run old task (before deprecation)
        task_old = TaskWithDeprecatedConfig.C(p=OldConfigForTask.C()).submit()
        task_old.wait()
        task_old_path = task_old.stdout().parent

        # Now deprecate the old config
        OldConfigForTask.__xpmtype__.deprecate()

        # Run fix_deprecated
        fix_deprecated(xp.workspace.path, fix=True, cleanup=False)

        # Verify symlink was created
        _check_symlink_paths(task_new, task_old_path)


class NewTask(Task):
    """New task for deprecation tests."""

    __xpmid__ = "test.deprecated.task.new"
    x: Param[int]

    def execute(self):
        pass


class OldTask(NewTask):
    """Old task without deprecate flag (for comparison)."""

    __xpmid__ = "test.deprecated.task.deprecated"


@deprecate
class DeprecatedTask(NewTask):
    """Deprecated task (legacy pattern)."""

    __xpmid__ = "test.deprecated.task.deprecated"


def test_task_deprecated_identifier():
    """Test that deprecated tasks have correct identifiers."""
    with TemporaryExperiment("deprecated_task", maxwait=20):
        task_new = NewTask.C(x=1).submit(run_mode=RunMode.DRY_RUN)
        task_old = OldTask.C(x=1).submit(run_mode=RunMode.DRY_RUN)
        task_deprecated = DeprecatedTask.C(x=1).submit(run_mode=RunMode.DRY_RUN)

        # Old and new should have different paths
        assert (
            task_new.stdout() != task_old.stdout()
        ), "Old and new path should be different"

        # Deprecated should have same path as new
        assert (
            task_new.stdout() == task_deprecated.stdout()
        ), "Deprecated path should be the same as non deprecated"


def test_task_deprecated_fix_deprecated():
    """Test fix_deprecated creates symlinks for deprecated tasks."""
    with TemporaryExperiment("deprecated_task_fix", maxwait=20) as xp:
        task_new = NewTask.C(x=1).submit(run_mode=RunMode.DRY_RUN)

        # Run old task (before deprecation)
        task_old = OldTask.C(x=1).submit()
        task_old.wait()
        task_old_path = task_old.stdout().parent

        # Deprecate and fix
        OldTask.__xpmtype__.deprecate()
        fix_deprecated(xp.workspace.path, fix=True, cleanup=False)

        # Verify symlink
        _check_symlink_paths(task_new, task_old_path)


# =============================================================================
# Extended tests for new deprecation mechanism with tasks
# =============================================================================


class NewTaskWithConvert(Task):
    """New task with different parameter structure."""

    __xpmid__ = "test.deprecated.task.convert.new"
    values: Param[List[int]]

    def execute(self):
        pass


@deprecate(NewTaskWithConvert)
class OldTaskWithConvert(Task):
    """Old task with single value, deprecated to new task with list."""

    __xpmid__ = "test.deprecated.task.convert.old"
    value: Param[int]

    def __convert__(self):
        return NewTaskWithConvert.C(values=[self.value])

    def execute(self):
        pass


def test_task_deprecated_with_convert_identifier():
    """Test deprecated task with __convert__ has correct identifier."""
    with TemporaryExperiment("deprecated_task_convert", maxwait=0):
        # Old task should compute identifier via __convert__
        task_old = OldTaskWithConvert.C(value=42).submit(run_mode=RunMode.DRY_RUN)
        task_new = NewTaskWithConvert.C(values=[42]).submit(run_mode=RunMode.DRY_RUN)

        # Identifiers should match (computed via __convert__)
        assert (
            task_old.stdout() == task_new.stdout()
        ), "Deprecated task should have same path as equivalent new task"


@deprecate(NewTaskWithConvert, replace=True)
class ReplacedTaskWithConvert(Task):
    """Old task that gets immediately replaced with new task."""

    __xpmid__ = "test.deprecated.task.replace.old"
    value: Param[int]

    def __convert__(self):
        return NewTaskWithConvert.C(values=[self.value])

    def execute(self):
        pass


def test_task_deprecated_replace_returns_new_type():
    """Test deprecated task with replace=True returns new task type."""
    with TemporaryExperiment("deprecated_task_replace", maxwait=0):
        # Creating old task should return new task
        result = ReplacedTaskWithConvert.C(value=42)

        # Should be NewTaskWithConvert type
        assert type(result).__name__ == "NewTaskWithConvert.XPMConfig"
        assert result.values == [42]

        # Submit should work with the new task
        task = result.submit(run_mode=RunMode.DRY_RUN)

        # Create equivalent new task directly
        task_new = NewTaskWithConvert.C(values=[42]).submit(run_mode=RunMode.DRY_RUN)

        # Paths should match
        assert task.stdout() == task_new.stdout()


# =============================================================================
# Attribute warning tests for replaced configs
# =============================================================================


def test_deprecate_replace_warns_on_old_attribute(caplog):
    """Test that setting attributes that existed on deprecated config warns."""

    class NewConfigForWarn(Config):
        __xpmid__ = "test.deprecate.replace.warn.new"
        values: Param[list[int]]

    @deprecate(NewConfigForWarn, replace=True)
    class OldConfigForWarn(Config):
        __xpmid__ = "test.deprecate.replace.warn.old"
        value: Param[int]
        extra: Param[str] = field(
            ignore_default="default"
        )  # This doesn't exist in NewConfigForWarn

        def __convert__(self):
            return NewConfigForWarn.C(values=[self.value])

    # Creating OldConfig returns NewConfig
    result = OldConfigForWarn.C(value=42)
    assert type(result).__name__ == "NewConfigForWarn.XPMConfig"

    # Setting an attribute that only exists on OldConfigForWarn should warn
    import logging

    with caplog.at_level(logging.WARNING):
        result.extra = "new_value"

    # Check that warning was logged
    assert any("extra" in record.message for record in caplog.records)
    assert any("deprecated" in record.message.lower() for record in caplog.records)

    # The attribute should NOT be set on the new config
    assert not hasattr(result, "extra") or getattr(result, "extra", None) is None
